#include "StdAfx.h"

#include <stdio.h>
#include <string.h>
#include "Ident.h"

#include "AbstractParseTree.h"

typedef struct context_t context_t, *context_p;
typedef struct tree_t tree_t, *tree_p;
typedef struct list_t list_t, *list_p;

struct tree_t
{   const char *type;
    union
    {   list_p parts;
		const char *ident;
        const char *str_value;
        int    int_value;
        double double_value;
        char   char_value;
        context_p context;
    } c;
    int line, column;
    unsigned long refcount;
};

struct list_t
{   
	list_p next;
    tree_p first;
};

/* Special strings, one for each terminal */
static const char *tt_ident = "identifier";
static const char *tt_str_value = "string value";
static const char *tt_int_value = "integer value";
static const char *tt_double_value = "double value";
static const char *tt_char_value = "char value";
static const char *tt_list = "list";

static const char *tt_opencontext = "<opencontext>";
static const char *tt_closecontext = "<closecontext>";

static tree_p old_trees = 0;
static list_p old_lists = 0;
static long alloced_trees = 0;

bool type_can_have_parts( const char *type )
{
	return    type != tt_ident 
           && type != tt_str_value
           && type != tt_int_value
           && type != tt_double_value
           && type != tt_char_value
           && type != tt_opencontext
           && type != tt_closecontext;
}

void free_list( list_p list );
void tree_release(tree_p tree)
{
    if (tree == 0)
        return;

	//printf("%8X -- (%d) ", tree, tree->refcount-1);
	//SContext::print(stdout);
	//printf("\n");

    alloced_trees--;
    tree->refcount--;

    if (tree->refcount == 0)
    {
        if (type_can_have_parts(tree->type))
        {   list_p list = tree->c.parts;

            while (list != 0)
            {   list_p next = list->next;
                free_list(list);
                list = next;
            }
        }

        tree->type = (char*)old_trees;
        old_trees = tree;
    }
}

void free_list( list_p list )
{	tree_release(list->first);
	list->next = old_lists;
	old_lists = list;
}


void AbstractParseTree::release()
{
	if (_tree != 0)
	{
		tree_release(_tree);
		_tree = 0;
	}
}


tree_p malloc_tree( void )
{   tree_p new_tree;

    if (old_trees)
    {   new_tree = old_trees;
        old_trees = (tree_p)old_trees->type;
    }
    else
        new_tree = new tree_t;

	new_tree->line = 0;
	new_tree->column = 0;
    new_tree->refcount = 1;
    alloced_trees++;

	//printf("%8X +A ", new_tree);
	//SContext::print(stdout);
	//printf("\n");

    return new_tree;
} 

list_p malloc_list( void )
{	list_p new_list;

    if (old_lists)
    {   new_list = old_lists;
        old_lists = old_lists->next;
    }
    else
        new_list = new list_t;
    new_list->next = 0;
    new_list->first = 0;
    
    return new_list;
}

void tree_assign(tree_p &d, tree_p s)
{
  tree_p old_d = d;

  d = s;
  if (d != 0)
  {
    d->refcount++;
    alloced_trees++;

	//printf("%8X +B ", d);
	//SContext::print(stdout);
	//printf("\n");
  }
  if (old_d != 0)
      tree_release(old_d);
}

AbstractParseTree::AbstractParseTree(const AbstractParseTree& lhs)
{
	_tree = lhs._tree;
	if (_tree != 0)
	{
		_tree->refcount++;
		alloced_trees++;

		//printf("%8X +C ", _tree);
		//SContext::print(stdout);
		//printf("\n");
	}
}

AbstractParseTree& AbstractParseTree::operator=(const AbstractParseTree& lhs)
{
	tree_assign(_tree, lhs._tree);
	return *this;
}

AbstractParseTree::AbstractParseTree(const AbstractParseTree::iterator& lhs)
{
	_tree = 0;
	tree_assign(_tree, lhs._list->first);
}

void AbstractParseTree::clear()
{
	if (_tree != 0)
	{
		tree_release(_tree);
		_tree = 0;
	}
}
void AbstractParseTree::attach(AbstractParseTree& treeRef)
{
	tree_p old_tree = _tree;
	_tree = treeRef._tree;
	treeRef._tree = 0;
	tree_release(old_tree);
}

void AbstractParseTree::create()
{   
	release();
	_tree = malloc_tree();
}


void AbstractParseTree::createIdent( const Ident ident )
{
	create();
    _tree->type = tt_ident;
    _tree->c.ident = ident.val(); 
}

void AbstractParseTree::createStringAtom( const char *str )
{
	create();
    _tree->type = tt_str_value;
    _tree->c.str_value = str; 
}

void AbstractParseTree::createDoubleAtom( double value )
{
	create();
    _tree->type = tt_double_value;
    _tree->c.double_value = value; 
}

void AbstractParseTree::createCharAtom( const char value )
{
	create();
    _tree->type = tt_char_value;
    _tree->c.char_value = value; 
}

void AbstractParseTree::createIntAtom( int value )
{
	create();
    _tree->type = tt_int_value;
    _tree->c.int_value = value; 
}

void AbstractParseTree::createList( void )
{
	create();
    _tree->type = tt_list;
    _tree->c.parts = 0;
}
    
void AbstractParseTree::createTree( const Ident name )
{
	create();
	ASSERT(!name.empty());
    _tree->type = name.val();
    _tree->c.parts = 0;
}

void AbstractParseTree::setTreeName( const Ident name )
{
	ASSERT(!name.empty());
	_tree->type = name.val();
}
   
void AbstractParseTree::insertChild( const AbstractParseTree& child )
{
    list_p r_list;

    r_list = malloc_list();
    tree_assign(r_list->first, child._tree);
    r_list->next = _tree->c.parts;
    _tree->c.parts = r_list;
} 

void AbstractParseTree::appendChild( const AbstractParseTree& child )
{
    list_p *r_list = &_tree->c.parts;

    while (*r_list != 0)
        r_list = &(*r_list)->next;

    *r_list = malloc_list();
    tree_assign((*r_list)->first, child._tree);
}

void AbstractParseTree::dropLastChild()
{
    list_p *r_list = &_tree->c.parts;

    if (*r_list == 0)
        return;

    while ((*r_list)->next != 0)
        r_list = &(*r_list)->next;

    free_list(*r_list);
    *r_list = 0;
} 

void AbstractParseTree::createOpenContext()
{
	create();
    _tree->type = tt_opencontext;
    _tree->c.parts = 0;
}

void AbstractParseTree::createCloseContext()
{
	create();
    _tree->type = tt_closecontext;
    _tree->c.parts = 0;
}

bool AbstractParseTree::isIdent() const
{
    return _tree != 0 && _tree->type == tt_ident;
}

bool AbstractParseTree::isIdent( const Ident ident ) const
{
    return    _tree != 0
           && _tree->type == tt_ident
           && _tree->c.ident == ident.val();
}

bool AbstractParseTree::equalIdent( const Ident ident ) const
{
    return _tree->c.ident == ident.val();
}

Ident AbstractParseTree::identName() const
{
	return _tree->c.ident;
}

bool AbstractParseTree::isString( ) const
{
    return _tree != 0 && _tree->type == tt_str_value;
}

bool AbstractParseTree::isString( const char *str ) const
{
    return    _tree != 0
           && _tree->type == tt_str_value
           && _tree->c.str_value == str;
}

const char* AbstractParseTree::stringValue() const
{
	return _tree->c.str_value;
}

bool AbstractParseTree::isInt() const
{
    return _tree != 0 && _tree->type == tt_int_value;
}

int AbstractParseTree::intValue() const
{
	return _tree->c.int_value;
} 

bool AbstractParseTree::isDouble() const
{
    return _tree != 0 && _tree->type == tt_double_value;
}

double AbstractParseTree::doubleValue() const
{
	return _tree->c.double_value;
}

bool AbstractParseTree::isChar() const
{
    return _tree != 0 && _tree->type == tt_char_value;
}

char AbstractParseTree::charValue() const
{
	return _tree->c.char_value;
}

bool AbstractParseTree::isList() const
{
    return _tree != 0 && _tree->type == tt_list;
}

bool AbstractParseTree::isTree( Ident name ) const
{
    return _tree && _tree->type == name.val();
}

bool AbstractParseTree::equalTree( Ident name ) const
{
    return _tree->type == name.val();
}

const char* AbstractParseTree::type() const
{
	return _tree->type;
}

int AbstractParseTree::nrParts() const
{
	list_p parts = _tree != 0 ? _tree->c.parts : 0;
	int nr = 0;
	
	for (; parts; parts = parts->next)
		nr++;
		
	return nr;
}	
	
AbstractParseTree AbstractParseTree::part( int i ) const
{
    list_p parts = _tree->c.parts;

    for (; parts && i > 1; parts = parts->next, i--);

	AbstractParseTree result;
	if (parts != 0)
		tree_assign(result._tree, parts->first);
	
	return result;
}

void AbstractParseTree::setLineColumn(int line, int column)
{
	if (_tree != 0)
	{
		_tree->line = line;
		_tree->column = column;
	}
}

int AbstractParseTree::line() const 
{
	return _tree->line;
}

int AbstractParseTree::column() const
{
	return _tree->column;
}

void AbstractParseTree::print( FILE *f, bool compact )
{
	static int print_tree_depth = 0;

    if (isEmpty())
    	fprintf(f, "[EMPTY]");
    else 
    {
    	if (line() != 0)
			fprintf(f, "<%d:%d>", line(), column());
 		if (isIdent())
    		fprintf(f, "%s", identName().val());
    	else if (isString())
    		fprintf(f, "\"%s\"", stringValue());
	    else if (isInt())
    	    fprintf(f, "%d", intValue());
    	else if (isDouble())
    	    fprintf(f, "%f", doubleValue());
    	else if (isChar())
    	    fprintf(f, "'%c'", charValue());
    	else if (type() == tt_opencontext)
    	{   fprintf(f, "{");
    		//print_context(f, tree->c.context, compact);
    	}
    	else if (type() == tt_closecontext)
    		fprintf(f, "}");
    	else
    	{    
        	fprintf(f, "%s(", type());
       
        	print_tree_depth += strlen(type()) + 1;

			bool first = true;
			for (AbstractParseTree::iterator child_it(*this); child_it.more(); child_it.next())
			{   
    			if (!first)
				{   if (compact)
						fprintf(f, ",");
					else 
						fprintf(f, ",\n%*s", print_tree_depth, "");
				}
				first = false;
				/* fprintf(f, "[%lx]", (longword)l); */
				AbstractParseTree(child_it).print(f, compact);
			}

			print_tree_depth -= strlen(type()) + 1;
        	fprintf(f, ")");
    	}
    }
    if (print_tree_depth == 0 && !compact)
    	fprintf(f, "\n");
};


AbstractParseTree::iterator::iterator(const AbstractParseTree& tree)
{
	_list = tree._tree != 0 ? tree._tree->c.parts : 0;
}
bool AbstractParseTree::iterator::more()
{
	return _list != 0;
}
void AbstractParseTree::iterator::next()
{
	_list = _list->next;
}




