#include "StdAfx.h"

#include "ScriptExecutor.h"

#include "Ident.h"
#include "AbstractParseTree.h"
#include "TextFileBuffer.h"
#include "Scanner.h"
#include "AbstractParser.h"
#include "IParser.h"
#include <math.h>
#include "Sample.h"
#include "SampleMultiDocTemplate.h"
#include "SampleDoc.h"
#include "SampleImageDoc.h"
#include "SampleImageView.h"
#include "gfl.h"
#include "MainFrm.h"
#include <string.h>

CCrystalTextView* currentView = 0;
void makeFullPath(const char* name, CString& fullPath)
{
	fullPath = CString(name);

	if (strchr(name, '\\') != 0)
		return;

	if (currentView == 0)
		return;

	CDocument* pDoc = currentView->GetDocument();
	CString path = pDoc->GetPathName();

	int pos = path.GetLength();
	for (; pos > 0; pos--)
		if (((LPCTSTR)path)[pos] == _T('\\'))
			break;
	fullPath = path.Left(pos+1) + fullPath;
}



static int roundToInt(double v) { return (int)floor(v + 0.5); }

class StringValue
{
public:
	StringValue() : _value(""), _allocated(false), _count(0) {}
	StringValue(const char *s) : _value((char*)s), _allocated(false), _count(0) {}
	StringValue(char *s, bool copy) : _allocated(copy), _count(0)
	{
		if (copy)
		{
			_value = new char[strlen(s)+1];
			strcpy(_value, s);
		}
		else
			_value = s;
	}
	StringValue(const StringValue& lhs, const StringValue& rhs) : _count(0)
	{
		int llhs = strlen(lhs.value());
		int lrhs = strlen(rhs.value());
		int len = llhs + lrhs;
		if (len == 0)
		{
			_allocated = false;
			_value = "";
		}
		else
		{
			_allocated = true;
			_value = new char[len + 1];
			strcpy(_value, lhs.value());
			strcpy(_value + llhs, rhs.value());
		}
	}
	~StringValue()
	{
		if (_allocated)
			delete _value;
	}
		
	const char* value() const
	{
		if (this == 0)
			return "";
		return _value; 
	}
	inline void claim() { _count++; }
	inline void release() { _count--; if (_count <= 0) delete this; }
private:
	char* _value;
	bool _allocated;
	long _count;
};

class ColorValue
{
public:
	ColorValue() : r(0.0), g(0.0), b(0.0), weight(0.0) {}
	ColorValue(double r_val, double g_val, double b_val, double weight_val = 1.0) : r(r_val), g(g_val), b(b_val), weight(weight_val) {}
	void clear() { r = g = b = weight = 0.0; }
	double r;
	double g;
	double b;
	double weight;
	const ColorValue operator+(const ColorValue& rhs) const
	{
		double combweight = weight + rhs.weight;
		if (fabs(combweight) < 0.00001)
			return ColorValue((r + rhs.r)/2, (g + rhs.g)/2, (b + rhs.b)/2, combweight);
		else
			return ColorValue((r * weight + rhs.r * rhs.weight)/combweight,
							  (g * weight + rhs.g * rhs.weight)/combweight,
							  (b * weight + rhs.b * rhs.weight)/combweight,
							  combweight);
	}
	const ColorValue operator-(const ColorValue& rhs) const
	{
		return (*this) + neg(rhs);
	}
	bool operator==(const ColorValue& rhs) const
	{
		return    roundToInt(r) == roundToInt(rhs.r)
			   && roundToInt(g) == roundToInt(rhs.g)
			   && roundToInt(b) == roundToInt(rhs.b);
	}
	bool operator!=(const ColorValue& rhs) const
	{
		return !(*this == rhs);
	}
private:
	static ColorValue neg(const ColorValue& rhs)
	{
		return ColorValue(rhs.r, rhs.g, rhs.b, -rhs.weight);
	}
};

class PointValue
{
public:
	PointValue() : x(0.0), y(0.0), weight(0.0) {}
	PointValue(double x_val, double y_val, double weight_val = 1.0) : x(x_val), y(y_val), weight(weight_val) {}
	void clear() { x = y = weight = 0.0; }
	double x;
	double y;
	double weight;
	const PointValue operator+(const PointValue& rhs) const
	{
		double combweight = weight + rhs.weight;
		if (fabs(combweight) < 0.00001)
			return PointValue((x + rhs.x)/2, (y + rhs.y)/2, combweight);
		else
			return PointValue((x * weight + rhs.x * rhs.weight)/combweight,
							  (y * weight + rhs.y * rhs.weight)/combweight,
							  combweight);
	}
	const PointValue operator*(double factor) const
	{
		return PointValue(x, y, factor * weight);
	}
	bool operator==(const PointValue& rhs) const
	{
		return    roundToInt(x) == roundToInt(rhs.x)
			   && roundToInt(y) == roundToInt(rhs.y);
	}
	bool operator!=(const PointValue& rhs) const
	{
		return !(*this == rhs);
	}
	double distance(const PointValue &rhs) const
	{ 
		double dx = x - rhs.x;
		double dy = y - rhs.y;
		return sqrt(dx*dx + dy*dy);
	}
};

class TextBufferAccessor
{
public:
	TextBufferAccessor() : _document(0), _textBuffer(0) {}
	void open(const char* name)
	{
		_document = 0;
		_textBuffer = 0;

		CSampleApp* app = (CSampleApp*)AfxGetApp();
		CSampleMultiDocTemplate* docTemplate = app->GetDocTemplate();

		// See if there is already a document with given name as file name or title
		{
			CString sName(name);

			POSITION pos = docTemplate->GetFirstDocPosition();
			while (pos != NULL)
			{
				CDocument* pDoc = docTemplate->GetNextDoc(pos);
				if (lstrcmpi(pDoc->GetPathName(), name) == 0 || pDoc->GetTitle() == sName)
				{
					CAbstractDocument* document = (CAbstractDocument*)pDoc;
					if (document->documentType() == 1)
					{
						_document = (CSampleDoc*)document;
						_textBuffer = &_document->m_xTextBuffer;
						_textView = _textBuffer->GetFirstView();
					}
					return;
				}
			}
		}

		CSampleDoc* pDocument = (CSampleDoc*)docTemplate->CreateNewDocument();

		BOOL bAutoDelete = pDocument->m_bAutoDelete;
		pDocument->m_bAutoDelete = FALSE;   // don't destroy if something goes wrong
		CFrameWnd* pFrame = docTemplate->CreateNewFrame(pDocument, NULL);
		pDocument->m_bAutoDelete = bAutoDelete;
		if (pFrame == NULL)
		{
			AfxMessageBox(AFX_IDP_FAILED_TO_CREATE_DOC);
			delete pDocument;       // explicit delete on error
			return;
		}
		ASSERT_VALID(pFrame);

		// open an existing document
		CWaitCursor wait;
		if (!GFLBitmap::isImageFormat(name))
		{
			pDocument->setSilent();
			if (!pDocument->OnOpenDocument(name))
			{
				pDocument->OnNewDocument();
			}
		}
		pDocument->SetPathName(name);

		docTemplate->InitialUpdateFrame(pFrame, pDocument, true);
		_document = pDocument;
		_textBuffer = &_document->m_xTextBuffer;
		_textView = _textBuffer->GetFirstView();
	}

	int nrLines()
	{
		if (_textBuffer == 0)
			return 0;

		return _textBuffer->GetLineCount();
	}

	char* line(int nr)
	{
		if (_textBuffer == 0 || nr < 0 || nr >= _textBuffer->GetLineCount())
			return "";

		return _textBuffer->GetLineChars(nr);
	}

	void clear()
	{
		if (_textBuffer == 0)
			return;

		int lastLine = _textBuffer->GetLineCount()-1;
		int lastPos = _textBuffer->GetLineLength(lastLine);
		if (lastLine > 0 || lastPos > 0)
			_textBuffer->DeleteText(_textView, 0, 0, lastLine, lastPos, CE_ACTION_DELETE); // [JRT]
	}

	void appendText(const char* s)
	{
		if (_textBuffer == 0)
			return;

		if (s == 0) // empty string
			return;

		int lastLine = _textBuffer->GetLineCount()-1;
		int lastPos = _textBuffer->GetLineLength(lastLine);
		int endLine;
		int endChar;
		_textBuffer->InsertText(_textView, lastLine, lastPos, s, endLine, endChar);
	}

private:
	CSampleDoc* _document;
	CCrystalTextBuffer* _textBuffer;
	CCrystalTextView* _textView;
};

class ImageAccessor
{
public:
	ImageAccessor() : _imageDocument(0), _gflBitmap(0), _changed(false) {}
	~ImageAccessor()
	{
		if (_changed && _imageDocument != 0)
		{
			POSITION pos = _imageDocument->GetFirstViewPosition();
			while (pos != NULL)
			{
				CSampleImageView* pView = (CSampleImageView*)_imageDocument->GetNextView(pos);
				pView->Invalidate();
			}
		}
	}

	void open(const char* name)
	{
		if (!GFLBitmap::initialized())
		{
			AfxMessageBox(_T("libgfl311.dll not loaded."));
			return;
		}

		_imageDocument = 0;

		CSampleApp* app = (CSampleApp*)AfxGetApp();
		CSampleMultiDocTemplate* docTemplate = app->GetDocTemplate();

		// See if there is already a document with given name as file name or title
		{
			CString sName(name);

			POSITION pos = docTemplate->GetFirstDocPosition();
			while (pos != NULL)
			{
				CDocument* pDoc = docTemplate->GetNextDoc(pos);
				if (lstrcmpi(pDoc->GetPathName(), name) == 0 || pDoc->GetTitle() == sName)
				{
					CAbstractDocument* document = (CAbstractDocument*)pDoc;
					if (document->documentType() == 2)
					{
						_imageDocument = (CSampleImageDoc*)document;
						_gflBitmap = &_imageDocument->m_gflBitmap;
						_changed = _gflBitmap->changed();
					}
					return;
				}
			}
		}

		CSampleImageDoc* pDocument = docTemplate->createDocImageObject();
		ASSERT_KINDOF(CDocument, pDocument);
		docTemplate->AddDocument(pDocument);

		BOOL bAutoDelete = pDocument->m_bAutoDelete;
		pDocument->m_bAutoDelete = FALSE;   // don't destroy if something goes wrong
		CFrameWnd* pFrame = docTemplate->CreateNewImageFrame(pDocument, NULL);
		pDocument->m_bAutoDelete = bAutoDelete;
		if (pFrame == NULL)
		{
			AfxMessageBox(AFX_IDP_FAILED_TO_CREATE_DOC);
			delete pDocument;       // explicit delete on error
			return;
		}
		ASSERT_VALID(pFrame);

		// open an existing document
		CWaitCursor wait;
		pDocument->SetPathName(name);
		if (GFLBitmap::isImageFormat(name))
		{
			pDocument->setSilent();
			CString fullPath;
			makeFullPath(name, fullPath);
			if (pDocument->OnOpenDocument(fullPath))
				pDocument->SetPathName(fullPath);
			else
				pDocument->OnNewDocument();
		}

		docTemplate->InitialUpdateFrame(pFrame, pDocument, true);
		_imageDocument = pDocument;
		_gflBitmap = &_imageDocument->m_gflBitmap;
		_changed = false;
	}

	void close()
	{
		if (!GFLBitmap::initialized())
		{
			AfxMessageBox(_T("libgfl311.dll not loaded."));
			return;
		}
		if (_gflBitmap == 0)
			return;
		_gflBitmap->clear();
		_imageDocument->OnCloseDocument();
		_imageDocument = 0;
	}

	void load()
	{
		if (!_changed) return;

		const char* name = _imageDocument->GetPathName();
		CString fullPath;
		makeFullPath(name, fullPath);
		_imageDocument->OnOpenDocument(fullPath);
		_changed = false;
	}

	int width()
	{
		return _gflBitmap != 0 ? _gflBitmap->width() : 0;
	}

	int height()
	{
		return _gflBitmap != 0 ? _gflBitmap->height() : 0;
	}

	void newImage(int width, int height)
	{
		if (_gflBitmap == 0)
			return;

		_gflBitmap->newImage(width, height);
		_changed = true;
	}

	ColorValue getColor(int x, int y)
	{
		if (_gflBitmap == 0)
			return ColorValue();

		double r, g, b;
		_gflBitmap->getColor(x, y, r, g, b);
		return ColorValue(r, g, b);
	}

	PointValue getPoint(int i)
	{
		if (1 <= i && i <= 9 && _imageDocument != 0 && _imageDocument->m_points[i-1].valid)
			return PointValue(_imageDocument->m_points[i-1].x, _imageDocument->m_points[i-1].y);
		return PointValue();
	}

	ColorValue getInterpolatedColor(double x, double y)
	{
		if (_gflBitmap == 0)
			return ColorValue();

		double r, g, b;
		if (!_gflBitmap->getInterpolatedColor(x, y, r, g, b))
			return ColorValue();

		return ColorValue(r, g, b);
	}

	bool setColor(int x, int y, const ColorValue& color)
	{
		if (_gflBitmap == 0)
			return false;

		if (color.weight == 0.0)
			return false;

		int r = roundToInt(color.r);
		int g = roundToInt(color.g);
		int b = roundToInt(color.b);
		if (!_gflBitmap->setColor(x, y, r, g, b))
			return false;

		_changed = true;
		return true;
	}

private:
	CSampleImageDoc* _imageDocument;
	GFLBitmap* _gflBitmap;
	bool _changed;
};

template <class T>
class Ptr
{
public:
	Ptr() : _ptr(0) {}
	Ptr(const Ptr& ptr) : _ptr(ptr._ptr) { if (_ptr != 0) _ptr->claim(); }
	Ptr(T* ptr) : _ptr(ptr) { if (_ptr != 0) _ptr->claim(); }
	~Ptr() { if (_ptr != 0) _ptr->release(); }
	const Ptr<T>&  operator=(Ptr<T> val)
	{
		T* oldPtr = _ptr;
		_ptr = val;
		if (_ptr != 0) _ptr->claim();
		if (oldPtr != 0) oldPtr->release();
		return *this;
	}
	const Ptr<T>& operator=(T* val)
	{
		T* oldPtr = _ptr;
		_ptr = val;
		if (_ptr != 0) _ptr->claim();
		if (oldPtr != 0) oldPtr->release();
		return *this;
	}
	operator T*(){ return _ptr; }
	T* operator*() { return _ptr; }
	T* operator->() { return _ptr; }
private:
	T* _ptr;
};

template <class T>
class PtrArray
{
public:
	PtrArray() : _ptrs(0) {}
	PtrArray(int size, T** ptrs)
	{
		_size = size;
		_ptrs = new T*[_size];
		for (int i = 0; i < _size; i++)
		{
			_ptrs[i] = ptrs[i];
			_ptrs[i]->claim();
		}
	}
	~PtrArray()
	{
		if (_ptrs != 0)
		{
			for (int i = 0; i < _size; i++)
				if (_ptrs[i] != 0)
					_ptrs[i]->release();
			delete _ptrs;
		}
	}
	void alloc(int size)
	{
		_size = size;
		if (_size > 0)
		{
			_ptrs = new T*[_size];
			for (int i = 0; i < _size; i++)
				_ptrs[i] = 0;
		}
	}
	void assign(int i, T* ptr) 
	{
		T* old = _ptrs[i];
		_ptrs[i] = ptr;
		if (ptr != 0) ptr->claim();
		if (old != 0) old->release();
	}
	int size() { return _size; }
	T* operator[](int i) { return _ptrs[i]; }
private:
	int _size;
	T** _ptrs;
};

//class StringPtr : public Ptr<StringValue>
//{
//};
typedef Ptr<StringValue> StringPtr;


//int total_nr_expr = 0;

class Expr
{
public:
	enum TypeEnum { typeNone, typeVoid,
					typeBool, typeBoolVar, typeBoolArrayVar,
					typeInt, typeIntVar, typeIntArrayVar,
					typeDouble, typeDoubleVar, typeDoubleArrayVar,
					typeString, typeStringVar, typeStringArrayVar,
					typeImage, typeImageVar,
					typeTextBuffer, typeTextBufferVar, 
					typeColor, typeColorVar, typeColorArrayVar,
					typePoint, typePointVar, typePointArrayVar };
	Expr() : _count(0)
	{
		//total_nr_expr++;
		//if (total_nr_expr == 761)
		//	TRACE("Lastv");
	}
	virtual TypeEnum type() { return typeNone; }
	virtual bool isRef() { return false; }
	virtual int nrDimensions() { return 0; }
	void claim() { _count++; }
	void release() { _count--; if (_count <= 0) delete this; }
	virtual ~Expr() {}
private:
	long _count;
};

class VoidExpr : public Expr
{
public:
	TypeEnum type() { return typeVoid; }
	virtual void eval() {}
};

class VarDecl : public VoidExpr
{
public:
	VarDecl(Ident ident, VarDecl* context) : _ident(ident), _context(context) {}
	VarDecl* find(Ident ident)
	{
		if (this == 0)
			return 0;
		if (ident == _ident)
			return this;
		return _context->find(ident);
	}
private:
	Ident _ident;
	VarDecl* _context;
};

class BoolExpr : public Expr
{
public:
	TypeEnum type() { return typeBool; }
	virtual bool eval() { return 0; }
};

class IntExpr : public Expr
{
public:
	TypeEnum type() { return typeInt; }
	virtual int eval() { return 0; }
};

class DoubleExpr : public Expr
{
public:
	TypeEnum type() { return typeDouble; }
	virtual double eval() { return 0; }
};

class StringExpr : public Expr
{
public:
	TypeEnum type() { return typeString; }
	virtual StringPtr eval() { return new StringValue(); }
};

class ColorExpr : public Expr
{
public:
	TypeEnum type() { return typeColor; }
	virtual const ColorValue eval() { return _value; }
protected:
	ColorValue _value;
};

class PointExpr : public Expr
{
public:
	TypeEnum type() { return typePoint; }
	virtual const PointValue eval() { return _value; }
protected:
	PointValue _value;
};

#define CASTTOVOID(F) \
class F##ToVoidExpr : public VoidExpr\
{\
public:\
	F##ToVoidExpr(F##Expr* expr) : _expr(expr) {}\
	void eval() { _expr->eval(); }\
private:\
	Ptr<F##Expr> _expr;\
};

//////////////// Generic classes /////////

class ArrayDecl
{
public:
	ArrayDecl(int nr_dim, IntExpr** dim_exprs) : _nr_dim(nr_dim), _dim_exprs(nr_dim, dim_exprs), _dims(new int[_nr_dim]), _alloc_size(0) {}
	virtual ~ArrayDecl()
	{
		delete _dims;
	}
	int nr_dim() { return _nr_dim; }
	int size(int dim) { return _dims[dim]; }
protected:
	long allocatedSize() { return _alloc_size; }
	bool allocatedSizeChanged()
	{
		long alloc_size = 1;
		for (int i = 0; i < _dim_exprs.size(); i++)
		{
			long size = _dim_exprs[i]->eval();
			if (size <= 0)
				size = 1;
			_dims[i] = size;
			alloc_size *= size;
		}

		if (_alloc_size != alloc_size)
		{
			_alloc_size = alloc_size;
			return true;
		}

		return false;
	}
private:
	int _nr_dim;
	PtrArray<IntExpr> _dim_exprs;
	int* _dims;
	long _alloc_size;
};

class ArrayDerefExpr
{
public:
	ArrayDerefExpr(int size, IntExpr** arg_exprs) : _arg_exprs(size, arg_exprs) {}
	long index(ArrayDecl& arrayDecl)
	{
		long total_size = 1;
		long offset = 0;
		for (int dim = 0; dim < arrayDecl.nr_dim(); dim++)
		{
			long size = arrayDecl.size(dim);
			long i = _arg_exprs[dim]->eval();
			if (i < 0)
				i = 0;
			if (i >= size)
				i = size - 1;
			offset += total_size * i;
			total_size *= size;
		}
		ASSERT(offset < total_size);
		return offset;
	}
private:
	PtrArray<IntExpr> _arg_exprs;
};


//////////////// BoolExpr ////////////////

class ConstBoolExpr : public BoolExpr
{
public:
	ConstBoolExpr(bool value) : _value(value) {}
	bool eval() { return _value; }
private:
	bool _value;
};

class NotBoolExpr : public BoolExpr
{
public: 
	NotBoolExpr(BoolExpr* rhs) : _rhs(rhs) {}
	bool eval() { return !_rhs->eval(); }
private:
	Ptr<BoolExpr> _rhs;
};

class IfBoolExpr : public BoolExpr
{
public:
	IfBoolExpr(BoolExpr* condExpr, BoolExpr* thenExpr, BoolExpr* elseExpr) : _condExpr(condExpr), _thenExpr(thenExpr), _elseExpr(elseExpr) {}
	bool eval() { return _condExpr->eval() ? _thenExpr->eval() : _elseExpr->eval(); }
private:
	Ptr<BoolExpr> _condExpr;
	Ptr<BoolExpr> _thenExpr;
	Ptr<BoolExpr> _elseExpr;
};

class DiadicBoolExpr : public BoolExpr
{
public:
	DiadicBoolExpr(BoolExpr* lhs, BoolExpr* rhs) : _lhs(lhs), _rhs(rhs) {}
protected:
	Ptr<BoolExpr> _lhs;
	Ptr<BoolExpr> _rhs;
};

#define CLASSDIADIC(C,T,RT,O) \
class C##T : public Diadic##T \
{\
public:\
	C##T(T* lhs, T* rhs) : Diadic##T(lhs, rhs) {}\
	RT eval() { return _lhs->eval() O _rhs->eval();}\
};

CLASSDIADIC(And, BoolExpr, bool, &&)
CLASSDIADIC(Or, BoolExpr, bool, ||)


class DiadicCompareIntExpr : public BoolExpr
{
public:
	DiadicCompareIntExpr(IntExpr* lhs, IntExpr* rhs) : _lhs(lhs), _rhs(rhs) {}
protected:
	Ptr<IntExpr> _lhs;
	Ptr<IntExpr> _rhs;
};

#define CLASSCOMPARE(C,T,O) \
class C##T##BoolExpr : public DiadicCompare##T##Expr \
{\
public:\
	C##T##BoolExpr(T##Expr* lhs, T##Expr* rhs) : DiadicCompare##T##Expr(lhs, rhs) {}\
	bool eval() { return _lhs->eval() O _rhs->eval();}\
};

CLASSCOMPARE(Equal, Int, ==)
CLASSCOMPARE(Unequal, Int, !=)
CLASSCOMPARE(Smaller, Int, <)
CLASSCOMPARE(SmallerOrEqual, Int, <=)
CLASSCOMPARE(Larger, Int, >)
CLASSCOMPARE(LargerOrEqual, Int, >=)

class DiadicCompareDoubleExpr : public BoolExpr
{
public:
	DiadicCompareDoubleExpr(DoubleExpr* lhs, DoubleExpr* rhs) : _lhs(lhs), _rhs(rhs) {}
protected:
	Ptr<DoubleExpr> _lhs;
	Ptr<DoubleExpr> _rhs;
};

CLASSCOMPARE(Equal, Double, ==)
CLASSCOMPARE(Unequal, Double, !=)
CLASSCOMPARE(Smaller, Double, <)
CLASSCOMPARE(SmallerOrEqual, Double, <=)
CLASSCOMPARE(Larger, Double, >)
CLASSCOMPARE(LargerOrEqual, Double, >=)

class VarBoolDecl : public VarDecl
{
public:
	VarBoolDecl(Ident ident, VarDecl* context, BoolExpr* lhs) : VarDecl(ident, context), _lhs(lhs) {}
	TypeEnum type() { return typeBoolVar; }
	void eval()
	{
		_var = _lhs != 0 ? _lhs->eval() : false;
	}
	bool& ref() { return _var; }
private:
	bool _var;
	Ptr<BoolExpr> _lhs;
};

class RefBoolExpr : public BoolExpr
{
public:
	bool isRef() { return true; }
	bool eval() { return *ref(); }
	void assign(bool value) { *ref() = value; }
	virtual bool* ref() = 0;
};

class VarBoolExpr : public RefBoolExpr
{
public:
	VarBoolExpr(bool& var) : _var(var) {}
	bool* ref() { return &_var; }
private:
	bool& _var;
};

class BoolArrayDecl : public VarDecl, public ArrayDecl
{
public:
	BoolArrayDecl(Ident ident, VarDecl* context, int nr_dim, IntExpr** dim_exprs) : VarDecl(ident, context), ArrayDecl(nr_dim, dim_exprs), _var(0) {}
	TypeEnum type() { return typeBoolArrayVar; }
	void eval()
	{
		if (allocatedSizeChanged())
		{
			delete _var;
			_var = new bool[allocatedSize()];
		}
		long allocated_size = allocatedSize();
		for (long j = 0; j < allocated_size; j++)
			_var[j] = 0;
	}
	bool* var(int offset) { return _var + offset; }
protected:
	bool* _var;
};

class ArrayVarBoolExpr : public Expr
{
public:
	ArrayVarBoolExpr(BoolArrayDecl& arrayDecl) : _arrayDecl(arrayDecl) {}
	TypeEnum type() { return typeBool; }
	int nrDimensions() { return _arrayDecl.nr_dim(); }
	BoolArrayDecl& arrayDecl() { return _arrayDecl; }
private:
	BoolArrayDecl& _arrayDecl;
};

class ArrayDerefBoolExpr : public RefBoolExpr, public ArrayDerefExpr
{
public:
	ArrayDerefBoolExpr(ArrayVarBoolExpr* target, int nr_args, IntExpr** arg_exprs) : _target(target), ArrayDerefExpr(nr_args, arg_exprs) {}
	bool* ref()
	{
		BoolArrayDecl& arrayDecl = _target->arrayDecl();
		_ref = arrayDecl.var(index(arrayDecl));
		return _ref;
	}
private:
	Ptr<ArrayVarBoolExpr> _target;
	bool* _ref;
};

CASTTOVOID(Bool)

//////////////// IntExpr ////////////////

class ConstIntExpr : public IntExpr
{
public:
	ConstIntExpr(int value) : _value(value) {}
	int eval() { return _value; }
private:
	int _value;
};

class NegateIntExpr : IntExpr
{
public: 
	NegateIntExpr(IntExpr* rhs) : _rhs(rhs) {}
	int eval() { return -_rhs->eval(); }
private:
	Ptr<IntExpr> _rhs;
};

class DiadicIntExpr : public IntExpr
{
public:
	DiadicIntExpr(IntExpr* lhs, IntExpr* rhs) : _lhs(lhs), _rhs(rhs) {}
protected:
	Ptr<IntExpr> _lhs;
	Ptr<IntExpr> _rhs;
};

class FloorIntExpr : public IntExpr
{
public:
	FloorIntExpr(DoubleExpr* arg) : _arg(arg) {}
	int eval() { return (int)floor(_arg->eval()); }
private:
	Ptr<DoubleExpr> _arg;
};

class RoundIntExpr : public IntExpr
{
public:
	RoundIntExpr(DoubleExpr* arg) : _arg(arg) {}
	int eval() { return (int)floor(_arg->eval() + 0.5); }
private:
	Ptr<DoubleExpr> _arg;
};

class AbsIntExpr : public IntExpr
{
public:
	AbsIntExpr(IntExpr* arg) : _arg(arg) {}
	int eval() { int value = _arg->eval(); return value < 0 ? -value : value; }
private:
	Ptr<IntExpr> _arg;
};

class RandomIntExpr : public IntExpr
{
public:
	RandomIntExpr(IntExpr* arg) : _arg(arg) {}
	int eval() { return rand() % _arg->eval(); }
private:
	Ptr<IntExpr> _arg;
};

CLASSDIADIC(Add, IntExpr, int, +)
CLASSDIADIC(Sub, IntExpr, int, -)
CLASSDIADIC(Times, IntExpr, int, *)
CLASSDIADIC(Div, IntExpr, int, /)
CLASSDIADIC(Mod, IntExpr, int, %)

class MaxIntExpr : public DiadicIntExpr
{
public:
	MaxIntExpr(IntExpr* lhs, IntExpr* rhs) : DiadicIntExpr(lhs, rhs) {}
	int eval()
	{
		int lhs = _lhs->eval();
		int rhs = _rhs->eval();
		return lhs > rhs ? lhs : rhs;
	}
};

class MinIntExpr : public DiadicIntExpr
{
public:
	MinIntExpr(IntExpr* lhs, IntExpr* rhs) : DiadicIntExpr(lhs, rhs) {}
	int eval()
	{
		int lhs = _lhs->eval();
		int rhs = _rhs->eval();
		return lhs < rhs ? lhs : rhs;
	}
};

class VarIntDecl : public VarDecl
{
public:
	VarIntDecl(Ident ident, VarDecl* context, IntExpr* lhs) : VarDecl(ident, context), _lhs(lhs) {}
	TypeEnum type() { return typeIntVar; }
	void eval()
	{
		_var = _lhs != 0 ? _lhs->eval() : false;
	}
	int& ref() { return _var; }
protected:
	int _var;
	Ptr<IntExpr> _lhs;
};

class RefIntExpr : public IntExpr
{
public:
	bool isRef() { return true; }
	int eval() { return *ref(); }
	void assign(int value) { *ref() = value; }
	void timesAssign(int value) { *ref() *= value; }
	void divAssign(int value) { *ref() /= value; }
	void modAssign(int value) { *ref() %= value; }
	void addAssign(int value) { *ref() += value; }
	void subAssign(int value) { *ref() -= value; }
	virtual int* ref() = 0;
};

class VarIntExpr : public RefIntExpr
{
public:
	VarIntExpr(int& var) : _var(var) {}
	int* ref() { return &_var; }
private:
	int& _var;
};

class PostIncExpr : public IntExpr
{
public:
	PostIncExpr(RefIntExpr* var) : _var(var) {}
	int eval()
	{
		int* ref = _var->ref();
		int result = *ref;
		(*ref)++;
		return result;
	}
private:
	Ptr<RefIntExpr> _var;
};

class PostDecExpr : public IntExpr
{
public:
	PostDecExpr(RefIntExpr* var) : _var(var) {}
	int eval()
	{
		int* ref = _var->ref();
		int result = *ref;
		(*ref)--;
		return result;
	}
private:
	Ptr<RefIntExpr> _var;
};

class PreIncExpr : public IntExpr
{
public:
	PreIncExpr(RefIntExpr* var) : _var(var) {}
	int eval()
	{
		int* ref = _var->ref();
		(*ref)++;
		return *ref;
	}
private:
	Ptr<RefIntExpr> _var;
};

class PreDecExpr : public IntExpr
{
public:
	PreDecExpr(RefIntExpr* var) : _var(var) {}
	int eval()
	{
		int* ref = _var->ref();
		(*ref)--;
		return *ref;
	}
private:
	Ptr<RefIntExpr> _var;
};

class IfIntExpr : public IntExpr
{
public:
	IfIntExpr(BoolExpr* condExpr, IntExpr* thenExpr, IntExpr* elseExpr) : _condExpr(condExpr), _thenExpr(thenExpr), _elseExpr(elseExpr) {}
	int eval() { return _condExpr->eval() ? _thenExpr->eval() : _elseExpr->eval(); }
private:
	Ptr<BoolExpr> _condExpr;
	Ptr<IntExpr> _thenExpr;
	Ptr<IntExpr> _elseExpr;
};

class IntArrayDecl : public VarDecl, public ArrayDecl
{
public:
	IntArrayDecl(Ident ident, VarDecl* context, int nr_dim, IntExpr** dim_exprs) : VarDecl(ident, context), ArrayDecl(nr_dim, dim_exprs), _var(0) {}
	TypeEnum type() { return typeIntArrayVar; }
	void eval()
	{
		if (allocatedSizeChanged())
		{
			delete _var;
			_var = new int[allocatedSize()];
		}
		long allocated_size = allocatedSize();
		for (long j = 0; j < allocated_size; j++)
			_var[j] = 0;
	}
	int* var(int offset) { return _var + offset; }
protected:
	int* _var;
};

class ArrayVarIntExpr : public Expr
{
public:
	ArrayVarIntExpr(IntArrayDecl& arrayDecl) : _arrayDecl(arrayDecl) {}
	TypeEnum type() { return typeInt; }
	int nrDimensions() { return _arrayDecl.nr_dim(); }
	IntArrayDecl& arrayDecl() { return _arrayDecl; }
private:
	IntArrayDecl& _arrayDecl;
};

class ArrayDerefIntExpr : public RefIntExpr, public ArrayDerefExpr
{
public:
	ArrayDerefIntExpr(ArrayVarIntExpr* target, int nr_args, IntExpr** arg_exprs) : _target(target), ArrayDerefExpr(nr_args, arg_exprs) {}
	int* ref()
	{
		IntArrayDecl& arrayDecl = _target->arrayDecl();
		_ref = arrayDecl.var(index(arrayDecl));
		return _ref;
	}
private:
	Ptr<ArrayVarIntExpr> _target;
	int* _ref;
};

CASTTOVOID(Int)

//////////////// DoubleExpr ////////////////

class ConstDoubleExpr : public DoubleExpr
{
public:
	ConstDoubleExpr(double value) : _value(value) {}
	double eval() { return _value; }
private:
	double _value;
};

class IntToDoubleExpr : public DoubleExpr
{
public: 
	IntToDoubleExpr(IntExpr* rhs) : _rhs(rhs) {}
	double eval() { return _rhs->eval(); }
private:
	Ptr<IntExpr> _rhs;
};

class AbsDoubleExpr : public DoubleExpr
{
public:
	AbsDoubleExpr(DoubleExpr* arg) : _arg(arg) {}
	double eval() { return fabs(_arg->eval()); }
private:
	Ptr<DoubleExpr> _arg;
};

class SqrtDoubleExpr : public DoubleExpr
{
public:
	SqrtDoubleExpr(DoubleExpr* arg) : _arg(arg) {}
	double eval() { return sqrt(_arg->eval()); }
private:
	Ptr<DoubleExpr> _arg;
};

class LogDoubleExpr : public DoubleExpr
{
public:
	LogDoubleExpr(DoubleExpr* arg) : _arg(arg) {}
	double eval() { return log(_arg->eval()); }
private:
	Ptr<DoubleExpr> _arg;
};

class SinDoubleExpr : public DoubleExpr
{
public:
	SinDoubleExpr(DoubleExpr* arg) : _arg(arg) {}
	double eval() { return sin(_arg->eval()); }
private:
	Ptr<DoubleExpr> _arg;
};

class CosDoubleExpr : public DoubleExpr
{
public:
	CosDoubleExpr(DoubleExpr* arg) : _arg(arg) {}
	double eval() { return cos(_arg->eval()); }
private:
	Ptr<DoubleExpr> _arg;
};

class Atan2DoubleExpr : public DoubleExpr
{
public:
	Atan2DoubleExpr(DoubleExpr* argx, DoubleExpr* argy) : _argx(argx), _argy(argy) {}
	double eval() { 
		return atan2(_argy->eval(), _argx->eval()); 
	}
private:
	Ptr<DoubleExpr> _argx;
	Ptr<DoubleExpr> _argy;
};

class RandomVarDoubleExpr : public DoubleExpr
{
public:
	RandomVarDoubleExpr(DoubleExpr* arg) : _arg(arg) {}
	double eval()
	{
		double range = _arg->eval();
		long v = rand();
		double x = (v/2)%10000 / 10000.0 * range;
		return v % 2 == 0 ? x : -x;
	}
private:
	Ptr<DoubleExpr> _arg;
};

class NegateDoubleExpr : public DoubleExpr
{
public: 
	NegateDoubleExpr(DoubleExpr* rhs) : _rhs(rhs) {}
	double eval() { return -_rhs->eval(); }
private:
	Ptr<DoubleExpr> _rhs;
};

class DiadicDoubleExpr : public DoubleExpr
{
public:
	DiadicDoubleExpr(DoubleExpr* lhs, DoubleExpr* rhs) : _lhs(lhs), _rhs(rhs) {}
protected:
	Ptr<DoubleExpr> _lhs;
	Ptr<DoubleExpr> _rhs;
};

CLASSDIADIC(Add, DoubleExpr, double, +)
CLASSDIADIC(Sub, DoubleExpr, double, -)
CLASSDIADIC(Times, DoubleExpr, double, *)
CLASSDIADIC(Div, DoubleExpr, double, /)

class MaxDoubleExpr : public DiadicDoubleExpr
{
public:
	MaxDoubleExpr(DoubleExpr* lhs, DoubleExpr* rhs) : DiadicDoubleExpr(lhs, rhs) {}
	double eval()
	{
		double lhs = _lhs->eval();
		double rhs = _rhs->eval();
		return lhs > rhs ? lhs : rhs;
	}
};

class MinDoubleExpr : public DiadicDoubleExpr
{
public:
	MinDoubleExpr(DoubleExpr* lhs, DoubleExpr* rhs) : DiadicDoubleExpr(lhs, rhs) {}
	double eval()
	{
		double lhs = _lhs->eval();
		double rhs = _rhs->eval();
		return lhs < rhs ? lhs : rhs;
	}
};

class VarDoubleDecl : public VarDecl
{
public:
	VarDoubleDecl(Ident ident, VarDecl* context, DoubleExpr* lhs) : VarDecl(ident, context), _lhs(lhs) {}
	TypeEnum type() { return typeDoubleVar; }
	void eval()
	{
		_var = _lhs != 0 ? _lhs->eval() : false;
	}
	double& ref() { return _var; }
private:
	double _var;
	Ptr<DoubleExpr> _lhs;
};

class RefDoubleExpr : public DoubleExpr
{
public:
	bool isRef() { return true; }
	double eval() { return *ref(); }
	void assign(double value) { *ref() = value; }
	void timesAssign(double value) { *ref() *= value; }
	void divAssign(double value) { *ref() /= value; }
	void addAssign(double value) { *ref() += value; }
	void subAssign(double value) { *ref() -= value; }
	virtual double* ref() = 0;
};

class SimpleVarDoubleExpr : public RefDoubleExpr
{
public:
	SimpleVarDoubleExpr(double& var) : _var(var) {}
	double* ref() { return &_var; }
private:
	double& _var;
};

class IfDoubleExpr : public DoubleExpr
{
public:
	IfDoubleExpr(BoolExpr* condExpr, DoubleExpr* thenExpr, DoubleExpr* elseExpr) : _condExpr(condExpr), _thenExpr(thenExpr), _elseExpr(elseExpr) {}
	double eval() { return _condExpr->eval() ? _thenExpr->eval() : _elseExpr->eval(); }
private:
	Ptr<BoolExpr> _condExpr;
	Ptr<DoubleExpr> _thenExpr;
	Ptr<DoubleExpr> _elseExpr;
};

class DoubleArrayDecl : public VarDecl, public ArrayDecl
{
public:
	DoubleArrayDecl(Ident ident, VarDecl* context, int nr_dim, IntExpr** dim_exprs) : VarDecl(ident, context), ArrayDecl(nr_dim, dim_exprs), _var(0) {}
	TypeEnum type() { return typeDoubleArrayVar; }
	void eval()
	{
		if (allocatedSizeChanged())
		{
			delete _var;
			_var = new double[allocatedSize()];
		}
		long allocated_size = allocatedSize();
		for (long j = 0; j < allocated_size; j++)
			_var[j] = 0;
	}
	double* var(int offset) { return _var + offset; }
protected:
	double* _var;
};

class ArrayVarDoubleExpr : public Expr
{
public:
	ArrayVarDoubleExpr(DoubleArrayDecl& arrayDecl) : _arrayDecl(arrayDecl) {}
	TypeEnum type() { return typeDouble; }
	int nrDimensions() { return _arrayDecl.nr_dim(); }
	DoubleArrayDecl& arrayDecl() { return _arrayDecl; }
private:
	DoubleArrayDecl& _arrayDecl;
};

class ArrayDerefDoubleExpr : public RefDoubleExpr, public ArrayDerefExpr
{
public:
	ArrayDerefDoubleExpr(ArrayVarDoubleExpr* target, int nr_args, IntExpr** arg_exprs) : _target(target), ArrayDerefExpr(nr_args, arg_exprs) {}
	double* ref()
	{
		DoubleArrayDecl& arrayDecl = _target->arrayDecl();
		_ref = arrayDecl.var(index(arrayDecl));
		return _ref;
	}
private:
	Ptr<ArrayVarDoubleExpr> _target;
	double* _ref;
};

CASTTOVOID(Double)

//////////////// StringExpr ////////////////

class ConstStringExpr : public StringExpr
{
public:
	ConstStringExpr(const char* value)
	{
		if (value != 0 && value != '\0')
			_value = new StringValue(value);
	}
	StringPtr eval() { return _value; }
private:
	StringPtr _value;
};

class IfStringExpr : public StringExpr
{
public:
	IfStringExpr(BoolExpr* condExpr, StringExpr* thenExpr, StringExpr* elseExpr) : _condExpr(condExpr), _thenExpr(thenExpr), _elseExpr(elseExpr) {}
	StringPtr eval() { return _condExpr->eval() ? _thenExpr->eval() : _elseExpr->eval(); }
private:
	Ptr<BoolExpr> _condExpr;
	Ptr<StringExpr> _thenExpr;
	Ptr<StringExpr> _elseExpr;
};

class AddStringExpr : public StringExpr
{
public:
	AddStringExpr(StringExpr* lhsExpr, StringExpr* rhsExpr) : _lhsExpr(lhsExpr), _rhsExpr(rhsExpr) {}
	StringPtr eval() { return new StringValue( _lhsExpr->eval()->value(), _rhsExpr->eval()->value() ); }
private:
	Ptr<StringExpr> _lhsExpr;
	Ptr<StringExpr> _rhsExpr;
};

class DiadicCompareStringExpr : public BoolExpr
{
public:
	DiadicCompareStringExpr(StringExpr* lhs, StringExpr* rhs) : _lhs(lhs), _rhs(rhs) {}
protected:
	Ptr<StringExpr> _lhs;
	Ptr<StringExpr> _rhs;
};

#define STRINGCOMPARE(C,O) \
class C##StringBoolExpr : public DiadicCompareStringExpr \
{\
public:\
	C##StringBoolExpr(StringExpr* lhs, StringExpr* rhs) : DiadicCompareStringExpr(lhs, rhs) {}\
	bool eval() { return strcmp(_lhs->eval()->value(), _rhs->eval()->value()) O 0;}\
};

STRINGCOMPARE(Equal, ==)
STRINGCOMPARE(Unequal, !=)
STRINGCOMPARE(Smaller, <)
STRINGCOMPARE(SmallerOrEqual, <=)
STRINGCOMPARE(Larger, >)
STRINGCOMPARE(LargerOrEqual, >=)

class StringAppender
{
public:
	StringAppender(char *&result) : _result(result), _pos(0), _len(0) {}
	~StringAppender()
	{
		add('\0');
		flush();
	}
	void add(const char c)
	{
		if (_pos == 1000)
			flush();
		_buffer[_pos++] = c;
	}
	void add(const char *s)
	{
		for (; *s != '\0'; s++)
			add(*s);
	}
private:
	void flush()
	{
		if (_pos > 0)
		{
			char *s = new char[_len + _pos];
			if (_len > 0)
				memcpy(s, _result, _len*sizeof(char));
			memcpy(s + _len, _buffer, _pos);
			_len += _pos;
			_pos = 0;
			delete _result;
			_result = s;
		}
	}
	char *&_result;
	int _pos;
	int _len;
	char _buffer[1000];
};

class FormatStringExpr : public StringExpr
{
public:
	FormatStringExpr(int nr_args, Expr** exprs) : _nr_args(nr_args), _exprs(nr_args, exprs) {}
	StringPtr eval()
	{
		char *result = 0;
		{
			StringAppender stringAppender(result);
			const char* s = ((StringExpr*)_exprs[0])->eval()->value();
			int expr_index = 0;
			while (*s != '\0')
				if (*s != '%')
					stringAppender.add(*s++);
				else if (s[1] == '%')
				{
					stringAppender.add('%');
					s += 2;
				}
				else
				{
					expr_index++;
					Expr *expr = _exprs[expr_index];
					char format[20];
					for (int i = 0; i < 18; i++, s++)
					{
						format[i] = *s;
						if (*s == '\0' || *s == 'd' || *s == 'f' || *s == 's')
						{
							format[i+1] = '\0';
							break;
						}
					}
					if (*s == '\0')
						stringAppender.add("<unknown format>");
					else if (expr_index >= _nr_args)
							stringAppender.add("<missing expression>");
					else if (*s == 'd')
					{
						if (expr->type() != Expr::typeInt)
							stringAppender.add("<no integer expression>");
						else
						{
							int intValue = ((IntExpr*)expr)->eval();
							char buffer[100];

							_snprintf(buffer, 99, format, intValue);
							buffer[99] = '\0';
							for (const char *r = buffer; *r != '\0'; r++)
								stringAppender.add(*r);
						}
					}
					else if (*s == 'f')
					{
						if (expr->type() != Expr::typeDouble)
							stringAppender.add("<no double expression>");
						else
						{
							double doubleValue = ((DoubleExpr*)expr)->eval();
							char buffer[100];

							_snprintf(buffer, 99, format, doubleValue);
							buffer[99] = '\0';
							for (const char *r = buffer; *r != '\0'; r++)
								stringAppender.add(*r);
						}
					}
					else if (*s == 's')
					{
						if (expr->type() != Expr::typeString)
							stringAppender.add("<no string expression>");
						else
							stringAppender.add(((StringExpr*)expr)->eval()->value());
					}
					if (*s != '\0')
						s++;
				}
		}
		return new StringValue(result, false);
	}
private:
	int _nr_args;
	PtrArray<Expr> _exprs;
};

class VarStringDecl : public VarDecl
{
public:
	VarStringDecl(Ident ident, VarDecl* context, StringExpr* lhs) : VarDecl(ident, context), _lhs(lhs) {}
	TypeEnum type() { return typeStringVar; }
	void eval()
	{
		if (_lhs != 0)
			_var = _lhs->eval();
	}
	StringPtr& ref() { return _var; }
private:
	StringPtr _var;
	Ptr<StringExpr> _lhs;
};

class RefStringExpr : public StringExpr
{
public:
	bool isRef() { return true; }
	StringPtr eval() { return *ref(); }
	void assign(const StringPtr& value) { *ref() = value; }
	virtual StringPtr* ref() = 0;
};

class VarStringExpr : public RefStringExpr
{
public:
	VarStringExpr(StringPtr& var) : _var(var) {}
	StringPtr* ref() { return &_var; }
private:
	StringPtr& _var;
};

class StringArrayDecl : public VarDecl, public ArrayDecl
{
public:
	StringArrayDecl(Ident ident, VarDecl* context, int nr_dim, IntExpr** dim_exprs) : VarDecl(ident, context), ArrayDecl(nr_dim, dim_exprs), _var(0) {}
	TypeEnum type() { return typeBoolArrayVar; }
	void eval()
	{
		if (allocatedSizeChanged())
		{
			delete _var;
			_var = new StringPtr[allocatedSize()];
		}
		long allocated_size = allocatedSize();
		for (long j = 0; j < allocated_size; j++)
			_var[j] = 0;
	}
	StringPtr* var(int offset) { return _var + offset; }
protected:
	StringPtr* _var;
};

class ArrayVarStringExpr : public Expr
{
public:
	ArrayVarStringExpr(StringArrayDecl& arrayDecl) : _arrayDecl(arrayDecl) {}
	TypeEnum type() { return typeString; }
	int nrDimensions() { return _arrayDecl.nr_dim(); }
	StringArrayDecl& arrayDecl() { return _arrayDecl; }
private:
	StringArrayDecl& _arrayDecl;
};

class ArrayDerefStringExpr : public RefStringExpr, public ArrayDerefExpr
{
public:
	ArrayDerefStringExpr(ArrayVarStringExpr* target, int nr_args, IntExpr** arg_exprs) : _target(target), ArrayDerefExpr(nr_args, arg_exprs) {}
	StringPtr* ref()
	{
		StringArrayDecl& arrayDecl = _target->arrayDecl();
		_ref = arrayDecl.var(index(arrayDecl));
		return _ref;
	}
private:
	Ptr<ArrayVarStringExpr> _target;
	StringPtr* _ref;
};

CASTTOVOID(String)

//////////////// TextBufferExpr ////////////////

class VarTextBufferDecl : public VarDecl
{
public:
	VarTextBufferDecl(Ident ident, VarDecl* context) : VarDecl(ident, context) {}
	TypeEnum type() { return typeTextBufferVar; }
	TextBufferAccessor& ref() { return _var; }
private:
	TextBufferAccessor _var;
};

class VarTextBufferExpr : public Expr
{
public:
	VarTextBufferExpr(TextBufferAccessor& var) : _var(var) {}
	TypeEnum type() { return typeTextBuffer; }
	bool isRef() { return true; }
	TextBufferAccessor& ref() { return _var; }
private:
	TextBufferAccessor& _var;
};

class TextBufferOpenMethodExpr : public VoidExpr
{
public:
	TextBufferOpenMethodExpr(VarTextBufferExpr* varTextBufferExpr, StringExpr* nameExpr) : _varTextBufferExpr(varTextBufferExpr), _nameExpr(nameExpr) {}
	void eval()
	{
		_varTextBufferExpr->ref().open(_nameExpr->eval()->value());
	}
private:
	Ptr<VarTextBufferExpr> _varTextBufferExpr;
	Ptr<StringExpr> _nameExpr;
};

class TextBufferNrLinesFieldExpr : public IntExpr
{
public:
	TextBufferNrLinesFieldExpr(VarTextBufferExpr* varTextBufferExpr) : _varTextBufferExpr(varTextBufferExpr) {}
	int eval()
	{
		return _varTextBufferExpr->ref().nrLines();
	}
private:
	Ptr<VarTextBufferExpr> _varTextBufferExpr;
};

class TextBufferLineMethodExpr : public StringExpr
{
public:
	TextBufferLineMethodExpr(VarTextBufferExpr* varTextBufferExpr, IntExpr* nrExpr) : _varTextBufferExpr(varTextBufferExpr), _nrExpr(nrExpr) {}
	StringPtr eval()
	{
		return new StringValue(_varTextBufferExpr->ref().line(_nrExpr->eval()), true);
	}
private:
	Ptr<VarTextBufferExpr> _varTextBufferExpr;
	Ptr<IntExpr> _nrExpr;
};

class TextBufferClearMethodExpr : public VoidExpr
{
public:
	TextBufferClearMethodExpr(VarTextBufferExpr* varTextBufferExpr) : _varTextBufferExpr(varTextBufferExpr) {}
	void eval()
	{
		_varTextBufferExpr->ref().clear();
	}
private:
	Ptr<VarTextBufferExpr> _varTextBufferExpr;
};

class TextBufferAppendMethodExpr : public VoidExpr
{
public:
	TextBufferAppendMethodExpr(VarTextBufferExpr* varTextBufferExpr, StringExpr* textExpr) : _varTextBufferExpr(varTextBufferExpr), _textExpr(textExpr) {}
	void eval()
	{
		const char *s = _textExpr->eval()->value();
		_varTextBufferExpr->ref().appendText(s);
	}
private:
	Ptr<VarTextBufferExpr> _varTextBufferExpr;
	Ptr<StringExpr> _textExpr;
};

//////////////// ImageExpr ////////////////

class VarImageDecl : public VarDecl
{
public:
	VarImageDecl(Ident ident, VarDecl* context) : VarDecl(ident, context) {}
	TypeEnum type() { return typeImageVar; }
	ImageAccessor& ref() { return _var; }
private:
	ImageAccessor _var;
};

class VarImageExpr : public Expr
{
public:
	VarImageExpr(ImageAccessor& var) : _var(var) {}
	TypeEnum type() { return typeImage; }
	bool isRef() { return true; }
	ImageAccessor& ref() { return _var; }
private:
	ImageAccessor& _var;
};

class ImageOpenMethodExpr : public VoidExpr
{
public:
	ImageOpenMethodExpr(VarImageExpr* varImageExpr, StringExpr* nameExpr) : _varImageExpr(varImageExpr), _nameExpr(nameExpr) {}
	void eval()
	{
		_varImageExpr->ref().open(_nameExpr->eval()->value());
	}
private:
	Ptr<VarImageExpr> _varImageExpr;
	Ptr<StringExpr> _nameExpr;
};

class ImageCloseMethodExpr : public VoidExpr
{
public:
	ImageCloseMethodExpr(VarImageExpr* varImageExpr) : _varImageExpr(varImageExpr) {}
	void eval()
	{
		_varImageExpr->ref().close();
	}
private:
	Ptr<VarImageExpr> _varImageExpr;
};

class ImageLoadMethodExpr : public VoidExpr
{
public:
	ImageLoadMethodExpr(VarImageExpr* varImageExpr) : _varImageExpr(varImageExpr) {}
	void eval()
	{
		_varImageExpr->ref().load();
	}
private:
	Ptr<VarImageExpr> _varImageExpr;
};

class ImageNewMethodExpr : public VoidExpr
{
public:
	ImageNewMethodExpr(VarImageExpr* varImageExpr, IntExpr* widthExpr, IntExpr* heightExpr) : _varImageExpr(varImageExpr), _widthExpr(widthExpr), _heightExpr(heightExpr) {}
	void eval()
	{
		_varImageExpr->ref().newImage(_widthExpr->eval(), _heightExpr->eval());
	}
private:
	Ptr<VarImageExpr> _varImageExpr;
	Ptr<IntExpr> _widthExpr;
	Ptr<IntExpr> _heightExpr;
};

class ImageGetColorMethodExpr : public ColorExpr
{
public:
	ImageGetColorMethodExpr(VarImageExpr* varImageExpr, IntExpr* xExpr, IntExpr* yExpr) : _varImageExpr(varImageExpr), _xExpr(xExpr), _yExpr(yExpr) {}
	const ColorValue eval()
	{
		_value = _varImageExpr->ref().getColor(_xExpr->eval(), _yExpr->eval());
		return _value;
	}
private:
	Ptr<VarImageExpr> _varImageExpr;
	Ptr<IntExpr> _xExpr;
	Ptr<IntExpr> _yExpr;
};

class ImageGetInterpolatedColorMethodExpr : public ColorExpr
{
public:
	ImageGetInterpolatedColorMethodExpr(VarImageExpr* varImageExpr, DoubleExpr* xExpr, DoubleExpr* yExpr) : _varImageExpr(varImageExpr), _xExpr(xExpr), _yExpr(yExpr) {}
	const ColorValue eval()
	{
		_value = _varImageExpr->ref().getInterpolatedColor(_xExpr->eval(), _yExpr->eval());
		return _value;
	}
private:
	Ptr<VarImageExpr> _varImageExpr;
	Ptr<DoubleExpr> _xExpr;
	Ptr<DoubleExpr> _yExpr;
};

class ImageGetInterpolatedPointColorMethodExpr : public ColorExpr
{
public:
	ImageGetInterpolatedPointColorMethodExpr(VarImageExpr* varImageExpr, PointExpr* pExpr) : _varImageExpr(varImageExpr), _pExpr(pExpr) {}
	const ColorValue eval()
	{
		PointValue point = _pExpr->eval();
		_value = _varImageExpr->ref().getInterpolatedColor(point.x, point.y);
		return _value;
	}
private:
	Ptr<VarImageExpr> _varImageExpr;
	Ptr<PointExpr> _pExpr;
};

class ImageSetColorMethodExpr : public VoidExpr
{
public:
	ImageSetColorMethodExpr(VarImageExpr* varImageExpr, IntExpr* xExpr, IntExpr* yExpr, ColorExpr* colorExpr) : _varImageExpr(varImageExpr), _xExpr(xExpr), _yExpr(yExpr), _colorExpr(colorExpr) {}
	void eval()
	{
		int x = _xExpr->eval();
		int y = _yExpr->eval();
		const ColorValue& color = _colorExpr->eval();
		_varImageExpr->ref().setColor(x, y, color);
	}
private:
	Ptr<VarImageExpr> _varImageExpr;
	ColorValue _value;
	Ptr<IntExpr> _xExpr;
	Ptr<IntExpr> _yExpr;
	Ptr<ColorExpr> _colorExpr;
};

class ImageGetPointMethodExpr : public PointExpr
{
public:
	ImageGetPointMethodExpr(VarImageExpr* varImageExpr, IntExpr* iExpr) : _varImageExpr(varImageExpr), _iExpr(iExpr) {}
	const PointValue eval()
	{
		_value = _varImageExpr->ref().getPoint(_iExpr->eval());
		return _value;
	}
private:
	Ptr<VarImageExpr> _varImageExpr;
	Ptr<IntExpr> _iExpr;
};

class ImageWidthFieldExpr : public IntExpr
{
public:
	ImageWidthFieldExpr(VarImageExpr* varImageExpr) : _varImageExpr(varImageExpr) {}
	int eval()
	{
		return _varImageExpr->ref().width();
	}
private:
	Ptr<VarImageExpr> _varImageExpr;
};

class ImageHeightFieldExpr : public IntExpr
{
public:
	ImageHeightFieldExpr(VarImageExpr* varImageExpr) : _varImageExpr(varImageExpr) {}
	int eval()
	{
		return _varImageExpr->ref().height();
	}
private:
	Ptr<VarImageExpr> _varImageExpr;
};

//////////////// ColorExpr ////////////////

class ConstructorColorExpr : public ColorExpr
{
public:
	ConstructorColorExpr(DoubleExpr* red_expr, DoubleExpr* green_expr, DoubleExpr* blue_expr) : _red_expr(red_expr), _green_expr(green_expr), _blue_expr(blue_expr) {}
	const ColorValue eval()
	{ 
		_value = ColorValue(_red_expr->eval(), _green_expr->eval(), _blue_expr->eval());
		return _value;
	}
private:
	Ptr<DoubleExpr> _red_expr;
	Ptr<DoubleExpr> _green_expr;
	Ptr<DoubleExpr> _blue_expr;
};

class IfColorExpr : public ColorExpr
{
public:
	IfColorExpr(BoolExpr* condExpr, ColorExpr* thenExpr, ColorExpr* elseExpr) : _condExpr(condExpr), _thenExpr(thenExpr), _elseExpr(elseExpr) {}
	const ColorValue eval() { return _condExpr->eval() ? _thenExpr->eval() : _elseExpr->eval(); }
private:
	Ptr<BoolExpr> _condExpr;
	Ptr<ColorExpr> _thenExpr;
	Ptr<ColorExpr> _elseExpr;
};

class DiadicCompareColorExpr : public BoolExpr
{
public:
	DiadicCompareColorExpr(ColorExpr* lhs, ColorExpr* rhs) : _lhs(lhs), _rhs(rhs) {}
protected:
	Ptr<ColorExpr> _lhs;
	Ptr<ColorExpr> _rhs;
};

CLASSCOMPARE(Equal, Color, ==)
CLASSCOMPARE(Unequal, Color, !=)

class DiadicColorExpr : public ColorExpr
{
public:
	DiadicColorExpr(ColorExpr* lhs, ColorExpr* rhs) : _lhs(lhs), _rhs(rhs) {}
protected:
	Ptr<ColorExpr> _lhs;
	Ptr<ColorExpr> _rhs;
};

CLASSDIADIC(Add, ColorExpr, const ColorValue, +)
CLASSDIADIC(Sub, ColorExpr, const ColorValue, -)


class VarColorDecl : public VarDecl
{
public:
	VarColorDecl(Ident ident, VarDecl* context, ColorExpr* lhs) : VarDecl(ident, context), _lhs(lhs) {}
	TypeEnum type() { return typeColorVar; }
	void eval()
	{
		_var = _lhs != 0 ? _lhs->eval() : ColorValue();
	}
	ColorValue& ref() { return _var; }
private:
	ColorValue _var;
	Ptr<ColorExpr> _lhs;
};

class RefColorExpr : public ColorExpr
{
public:
	bool isRef() { return true; }
	const ColorValue eval() { return ref(); }
	void assign(const ColorValue& value) { ref() = value; }
	void timesAssign(double value) { ref().weight *= value; }
	void addAssign(const ColorValue& value) { ref() = ref() + value; }
	void subAssign(const ColorValue& value) { ref() = ref() - value; }
	virtual ColorValue& ref() = 0;
};

class VarColorExpr : public RefColorExpr
{
public:
	VarColorExpr(ColorValue& var) : _var(var) {}
	ColorValue& ref() { return _var; }
private:
	ColorValue& _var;
};

#define COLORFIELDEXPR(C,V) \
class C : public RefDoubleExpr\
{\
public:\
	C(VarColorExpr* varColorExpr) : _varColorExpr(varColorExpr) {}\
	double* ref() { return &_varColorExpr->ref().V; }\
private:\
	Ptr<VarColorExpr> _varColorExpr;\
};

COLORFIELDEXPR(ColorRedFieldExpr, r)
COLORFIELDEXPR(ColorGreenFieldExpr, g)
COLORFIELDEXPR(ColorBlueFieldExpr, b)

class ColorArrayDecl : public VarDecl, public ArrayDecl
{
public:
	ColorArrayDecl(Ident ident, VarDecl* context, int nr_dim, IntExpr** dim_exprs) : VarDecl(ident, context), ArrayDecl(nr_dim, dim_exprs), _var(0) {}
	TypeEnum type() { return typeColorArrayVar; }
	void eval()
	{
		if (allocatedSizeChanged())
		{
			delete _var;
			_var = new ColorValue[allocatedSize()];
		}
		else
		{
			long allocated_size = allocatedSize();
			for (long j = 0; j < allocated_size; j++)
				_var[j].clear();
		}
	}
	ColorValue& var(int offset) { return _var[offset]; }
protected:
	ColorValue* _var;
};

class ArrayVarColorExpr : public Expr
{
public:
	ArrayVarColorExpr(ColorArrayDecl& arrayDecl) : _arrayDecl(arrayDecl) {}
	TypeEnum type() { return typeColor; }
	int nrDimensions() { return _arrayDecl.nr_dim(); }
	ColorArrayDecl& arrayDecl() { return _arrayDecl; }
private:
	ColorArrayDecl& _arrayDecl;
};

class ArrayDerefColorExpr : public RefColorExpr, public ArrayDerefExpr
{
public:
	ArrayDerefColorExpr(ArrayVarColorExpr* target, int nr_args, IntExpr** arg_exprs) : _target(target), ArrayDerefExpr(nr_args, arg_exprs) {}
	ColorValue& ref()
	{
		ColorArrayDecl& arrayDecl = _target->arrayDecl();
		_ref = &arrayDecl.var(index(arrayDecl));
		return *(ColorValue*)_ref;
	}
private:
	Ptr<ArrayVarColorExpr> _target;
	ColorValue* _ref;
};

CASTTOVOID(Color)

//////////////// PointExpr ////////////////

class ConstructorPointExpr : public PointExpr
{
public:
	ConstructorPointExpr(DoubleExpr* x_expr, DoubleExpr* y_expr) : _x_expr(x_expr), _y_expr(y_expr) {}
	const PointValue eval()
	{ 
		_value = PointValue(_x_expr->eval(), _y_expr->eval());
		return _value;
	}
private:
	Ptr<DoubleExpr> _x_expr;
	Ptr<DoubleExpr> _y_expr;
};

class IfPointExpr : public PointExpr
{
public:
	IfPointExpr(BoolExpr* condExpr, PointExpr* thenExpr, PointExpr* elseExpr) : _condExpr(condExpr), _thenExpr(thenExpr), _elseExpr(elseExpr) {}
	const PointValue eval() { return _condExpr->eval() ? _thenExpr->eval() : _elseExpr->eval(); }
private:
	Ptr<BoolExpr> _condExpr;
	Ptr<PointExpr> _thenExpr;
	Ptr<PointExpr> _elseExpr;
};

class DiadicComparePointExpr : public BoolExpr
{
public:
	DiadicComparePointExpr(PointExpr* lhs, PointExpr* rhs) : _lhs(lhs), _rhs(rhs) {}
protected:
	Ptr<PointExpr> _lhs;
	Ptr<PointExpr> _rhs;
};

CLASSCOMPARE(Equal, Point, ==)
CLASSCOMPARE(Unequal, Point, !=)

class DiadicPointExpr : public PointExpr
{
public:
	DiadicPointExpr(PointExpr* lhs, PointExpr* rhs) : _lhs(lhs), _rhs(rhs) {}
protected:
	Ptr<PointExpr> _lhs;
	Ptr<PointExpr> _rhs;
};

CLASSDIADIC(Add, PointExpr, const PointValue, +)

class TimesPointExpr : public PointExpr
{
public:
	TimesPointExpr(DoubleExpr* factor, PointExpr* point) : _factor(factor), _point(point) {}
	const PointValue eval()
	{
		return _point->eval() * _factor->eval();		
	}
protected:
	Ptr<DoubleExpr> _factor;
	Ptr<PointExpr> _point;
};

class VarPointDecl : public VarDecl
{
public:
	VarPointDecl(Ident ident, VarDecl* context, PointExpr* lhs) : VarDecl(ident, context), _lhs(lhs) {}
	TypeEnum type() { return typePointVar; }
	void eval()
	{
		_var = _lhs != 0 ? _lhs->eval() : PointValue();
	}
	PointValue& ref() { return _var; }
private:
	PointValue _var;
	Ptr<PointExpr> _lhs;
};

class RefPointExpr : public PointExpr
{
public:
	bool isRef() { return true; }
	const PointValue eval() { return ref(); }
	void assign(const PointValue& value) { ref() = value; }
	void timesAssign(double value) { ref().weight *= value; }
	void addAssign(const PointValue& value) { ref() = ref() + value; }
	virtual PointValue& ref() = 0;
};

class VarPointExpr : public RefPointExpr
{
public:
	VarPointExpr(PointValue& var) : _var(var) {}
	PointValue& ref() { return _var; }
private:
	PointValue& _var;
};

#define POINTFIELDEXPR(C,V) \
class C : public RefDoubleExpr\
{\
public:\
	C(VarPointExpr* varPointExpr) : _varPointExpr(varPointExpr) {}\
	double* ref() { return &_varPointExpr->ref().V; }\
private:\
	Ptr<VarPointExpr> _varPointExpr;\
};

POINTFIELDEXPR(PointXFieldExpr, x)
POINTFIELDEXPR(PointYFieldExpr, y)

class PointArrayDecl : public VarDecl, public ArrayDecl
{
public:
	PointArrayDecl(Ident ident, VarDecl* context, int nr_dim, IntExpr** dim_exprs) : VarDecl(ident, context), ArrayDecl(nr_dim, dim_exprs), _var(0) {}
	TypeEnum type() { return typePointArrayVar; }
	void eval()
	{
		if (allocatedSizeChanged())
		{
			delete _var;
			_var = new PointValue[allocatedSize()];
		}
		else
		{
			long allocated_size = allocatedSize();
			for (long j = 0; j < allocated_size; j++)
				_var[j].clear();
		}
	}
	PointValue& var(int offset) { return _var[offset]; }
protected:
	PointValue* _var;
};

class ArrayVarPointExpr : public Expr
{
public:
	ArrayVarPointExpr(PointArrayDecl& arrayDecl) : _arrayDecl(arrayDecl) {}
	TypeEnum type() { return typePoint; }
	int nrDimensions() { return _arrayDecl.nr_dim(); }
	PointArrayDecl& arrayDecl() { return _arrayDecl; }
private:
	PointArrayDecl& _arrayDecl;
};

class ArrayDerefPointExpr : public RefPointExpr, public ArrayDerefExpr
{
public:
	ArrayDerefPointExpr(ArrayVarPointExpr* target, int nr_args, IntExpr** arg_exprs) : _target(target), ArrayDerefExpr(nr_args, arg_exprs) {}
	PointValue& ref()
	{
		PointArrayDecl& arrayDecl = _target->arrayDecl();
		_ref = &arrayDecl.var(index(arrayDecl));
		return *(PointValue*)_ref;
	}
private:
	Ptr<ArrayVarPointExpr> _target;
	PointValue* _ref;
};

CASTTOVOID(Point)

class PointDistanceDoubleExpr : public DoubleExpr
{
public:
	PointDistanceDoubleExpr(PointExpr* p1_expr, PointExpr* p2_expr) : _p1_expr(p1_expr), _p2_expr(p2_expr) {}
	double eval() { return _p1_expr->eval().distance(_p2_expr->eval()); }
private:
	Ptr<PointExpr> _p1_expr;
	Ptr<PointExpr> _p2_expr;
};

//////////////// VoidExpr ////////////////

class StatementsExpr : public VoidExpr
{
public:
	StatementsExpr(int n) : _n(n)
	{
		_statements.alloc(n);
	}
	void set(int i, VoidExpr* expr)
	{
		if (0 <= i && i < _n)
			_statements.assign(i, expr);
	}
	void eval()
	{
		for (int i = 0; i < _n; i++)
			if (_statements[i] != 0)
				_statements[i]->eval();
	}
private:
	int _n;
	PtrArray<VoidExpr> _statements;
};

class IfExpr : public VoidExpr
{
public:
	IfExpr(BoolExpr* condition, VoidExpr* then_part, VoidExpr* else_part) : _condition(condition), _then_part(then_part), _else_part(else_part) {}
	void eval()
	{
		if (_condition->eval())
			_then_part->eval();
		else if (_else_part != 0)
			_else_part->eval();
	}
private:
	Ptr<BoolExpr> _condition;
	Ptr<VoidExpr> _then_part;
	Ptr<VoidExpr> _else_part;
};

class ForIntExpr : public VoidExpr
{
public:
	ForIntExpr(VarIntDecl* varExpr, IntExpr* fromExpr, IntExpr* toExpr, VoidExpr* statement) : _varExpr(varExpr), _fromExpr(fromExpr), _toExpr(toExpr), _statement(statement) {}
	void setStatement(VoidExpr* statement) { _statement = statement; }
	void eval()
	{
		int& var = _varExpr->ref();
		var = _fromExpr->eval();
		int to = _toExpr->eval();
		for (; var < to; var++)
		{
			_statement->eval();
		}
	}
private:
	Ptr<VarIntDecl> _varExpr;
	Ptr<IntExpr> _fromExpr;
	Ptr<IntExpr> _toExpr;
	Ptr<VoidExpr> _statement;
};

class WhileExpr : public VoidExpr
{
public:
	WhileExpr(BoolExpr* condition, VoidExpr* body) : _condition(condition), _body(body) {}
	void eval()
	{
		while (_condition->eval())
			_body->eval();
	}
private:
	Ptr<BoolExpr> _condition;
	Ptr<VoidExpr> _body;
};

#define ASSIGNMENTEXPR2(C,T,S,O) \
class C : public VoidExpr\
{\
public:\
	C(Ref##T##Expr* lhs, S##Expr* rhs) : _lhs(lhs), _rhs(rhs) {}\
	void eval() { _lhs->O(_rhs->eval()); }\
private:\
	Ptr<Ref##T##Expr> _lhs;\
	Ptr<S##Expr> _rhs;\
};
#define ASSIGNMENTEXPR(C,T,O) ASSIGNMENTEXPR2(C,T,T,O)

ASSIGNMENTEXPR(BoolAssignmentExpr, Bool, assign)

ASSIGNMENTEXPR(IntAssignmentExpr, Int, assign)
ASSIGNMENTEXPR(IntTimesAssignmentExpr, Int, timesAssign)
ASSIGNMENTEXPR(IntDivAssignmentExpr, Int, divAssign)
ASSIGNMENTEXPR(IntModAssignmentExpr, Int, modAssign)
ASSIGNMENTEXPR(IntAddAssignmentExpr, Int, addAssign)
ASSIGNMENTEXPR(IntSubAssignmentExpr, Int, subAssign)

ASSIGNMENTEXPR(DoubleAssignmentExpr, Double, assign)
ASSIGNMENTEXPR(DoubleTimesAssignmentExpr, Double, timesAssign)
ASSIGNMENTEXPR(DoubleDivAssignmentExpr, Double, divAssign)
ASSIGNMENTEXPR(DoubleAddAssignmentExpr, Double, addAssign)
ASSIGNMENTEXPR(DoubleSubAssignmentExpr, Double, subAssign)

ASSIGNMENTEXPR(StringAssignmentExpr, String, assign)

ASSIGNMENTEXPR(ColorAssignmentExpr, Color, assign)
ASSIGNMENTEXPR2(ColorTimesAssignmentExpr, Color, Double, timesAssign)
ASSIGNMENTEXPR(ColorAddAssignmentExpr, Color, addAssign)
ASSIGNMENTEXPR(ColorSubAssignmentExpr, Color, subAssign)

ASSIGNMENTEXPR(PointAssignmentExpr, Point, assign)
ASSIGNMENTEXPR2(PointTimesAssignmentExpr, Point, Double, timesAssign)
ASSIGNMENTEXPR(PointAddAssignmentExpr, Point, addAssign)

//////////////// Context ////////////////
/*
class Context
{
public:
	Context(Ident ident, VoidExpr* decl_expr, Context* next) : _ident(ident), _decl_expr(decl_expr), _next(next) {}
	VoidExpr* find(Ident ident)
	{
		if (this == 0)
			return 0;
		if (ident == _ident)
			return _decl_expr;
		return _next->find(ident);
	}
private:
	Ident _ident;
	Ptr<VoidExpr> _decl_expr;
	Context* _next;
};
*/

void setCursorAt(int line, int column)
{
	CPoint point(column, line);
	if (currentView == 0 || !currentView->SaveCursorPos(point))
		return;
	currentView->EnsureVisible(point);
	currentView->SetCursorPos(point);
}

bool reported_error = false;

void error(int line, int column, const char* msg)
{
	if (reported_error)
		return;
	reported_error = true;

	setCursorAt(line-1, column-1);

	CString prompt;
	prompt.Format("%d,%d: %s", line, column, msg);
	AfxMessageBox(prompt);
	CMainFrame* pMainFrame = (CMainFrame*) AfxGetApp()->m_pMainWnd;
	//if (pMainFrame != 0)
	//	pMainFrame->UpdateStatus(msg);
}

void error(AbstractParseTree tree, const char* msg)
{
	if (tree.isEmpty())
		error(0, 0, msg);
	else
		error(tree.line(), tree.column(), msg);
}

bool assertBoolType(Expr* expr, AbstractParseTree tree)
{
	if (expr != 0 && expr->nrDimensions() == 0  && expr->type() == Expr::typeBool)
		return true;
	error(tree, "Expect bool value");
	return false;
}

bool assertIntType(Expr* expr, AbstractParseTree tree)
{
	if (expr != 0 && expr->nrDimensions() == 0  && expr->type() == Expr::typeInt)
		return true;
	error(tree, "Expect int value");
	return false;
}

bool assertDoubleIntType(Expr* expr, AbstractParseTree tree)
{
	if (expr != 0 && expr->nrDimensions() == 0  && (expr->type() == Expr::typeInt || expr->type() == Expr::typeDouble))
		return true;
	error(tree, "Expect int or double value");
	return false;
}

bool assertDoubleIntColorPointStringType(Expr* expr, AbstractParseTree tree)
{
	if (expr != 0 && expr->nrDimensions() == 0 && (expr->type() == Expr::typeInt || expr->type() == Expr::typeDouble || expr->type() == Expr::typeColor || expr->type() == Expr::typePoint || expr->type() == Expr::typeString))
		return true;
	error(tree, "Expect int, double, or color value");
	return false;
}

bool assertBoolDoubleIntColorPointStringType(Expr* expr, AbstractParseTree tree)
{
	if (expr != 0 && expr->nrDimensions() == 0 && (expr->type() == Expr::typeBool || expr->type() == Expr::typeInt || expr->type() == Expr::typeDouble || expr->type() == Expr::typeColor || expr->type() == Expr::typePoint || expr->type() == Expr::typeString))
		return true;
	error(tree, "Expect bool, int, double, or color value");
	return false;
}

bool assertDoubleType(Expr* expr, AbstractParseTree tree)
{
	if (expr != 0 && expr->nrDimensions() == 0  && expr->type() == Expr::typeDouble)
		return true;
	error(tree, "Expect int or double value");
	return false;
}

bool assertStringType(Expr* expr, AbstractParseTree tree)
{
	if (expr != 0 && expr->nrDimensions() == 0  && expr->type() == Expr::typeString)
		return true;
	error(tree, "Expect string value");
	return false;
}

bool assertColorType(Expr* expr, AbstractParseTree tree)
{
	if (expr != 0 && expr->nrDimensions() == 0  && expr->type() == Expr::typeColor)
		return true;
	error(tree, "Expect color value");
	return false;
}

bool assertPointType(Expr* expr, AbstractParseTree tree)
{
	if (expr != 0 && expr->nrDimensions() == 0  && expr->type() == Expr::typePoint)
		return true;
	error(tree, "Expect point value");
	return false;
}

bool assertRefIntType(Expr* expr, AbstractParseTree tree)
{
	if (expr != 0 && expr->type() == Expr::typeInt && expr->isRef())
		return true;
	error(tree, "Expect integer reference");
	return false;
}

VoidExpr* toVoidExpr(Expr* expr, AbstractParseTree tree)
{
	if (expr == 0 || expr->type() == Expr::typeVoid)
		return (VoidExpr*)expr;
	if (expr->type() == Expr::typeBool)
		return new BoolToVoidExpr((BoolExpr*)expr);
	if (expr->type() == Expr::typeInt)
		return new IntToVoidExpr((IntExpr*)expr);
	if (expr->type() == Expr::typeDouble)
		return new DoubleToVoidExpr((DoubleExpr*)expr);
	if (expr->type() == Expr::typeString)
		return new StringToVoidExpr((StringExpr*)expr);
	if (expr->type() == Expr::typeColor)
		return new ColorToVoidExpr((ColorExpr*)expr);
	if (expr->type() == Expr::typePoint)
		return new PointToVoidExpr((PointExpr*)expr);
	error(tree, "Unknown type");
	return new VoidExpr;
}

Expr** compile_list(AbstractParseTree tree, int &nr_args, VarDecl* &context);

Expr* compile(AbstractParseTree tree, VarDecl* &context)
{
	if (tree.isInt())
		return new ConstIntExpr(tree.intValue());
	if (tree.isDouble())
		return new ConstDoubleExpr(tree.doubleValue());
	if (tree.isString())
		return new ConstStringExpr(tree.stringValue());
	if (tree.isIdent())
	{
		Expr* expr = context->find(tree.identName());
		if (expr == 0) { error(tree, "Variable not declared"); return new Expr(); }
		if (expr->type() == Expr::typeBoolVar)
		{
			VarBoolDecl* varBoolDecl = (VarBoolDecl*)expr;
			return new VarBoolExpr(varBoolDecl->ref());
		}
		if (expr->type() == Expr::typeIntVar)
		{
			VarIntDecl* varIntDecl = (VarIntDecl*)expr;
			return new VarIntExpr(varIntDecl->ref());
		}
		if (expr->type() == Expr::typeDoubleVar)
		{
			VarDoubleDecl* varDoubleDecl = (VarDoubleDecl*)expr;
			return new SimpleVarDoubleExpr(varDoubleDecl->ref());
		}
		if (expr->type() == Expr::typeStringVar)
		{
			VarStringDecl* varStringDecl = (VarStringDecl*)expr;
			return new VarStringExpr(varStringDecl->ref());
		}
		if (expr->type() == Expr::typeTextBufferVar)
		{
			VarTextBufferDecl* varTextBufferDecl = (VarTextBufferDecl*)expr;
			return new VarTextBufferExpr(varTextBufferDecl->ref());
		}
		if (expr->type() == Expr::typeImageVar)
		{
			VarImageDecl* varImageDecl = (VarImageDecl*)expr;
			return new VarImageExpr(varImageDecl->ref());
		}
		if (expr->type() == Expr::typeColorVar)
		{
			VarColorDecl* varColorDecl = (VarColorDecl*)expr;
			return new VarColorExpr(varColorDecl->ref());
		}
		if (expr->type() == Expr::typePointVar)
		{
			VarPointDecl* varPointDecl = (VarPointDecl*)expr;
			return new VarPointExpr(varPointDecl->ref());
		}
		if (expr->type() == Expr::typeBoolArrayVar)
		{
			BoolArrayDecl* boolArrayDecl = (BoolArrayDecl*)expr;
			return new ArrayVarBoolExpr(*boolArrayDecl);
		}
		if (expr->type() == Expr::typeIntArrayVar)
		{
			IntArrayDecl* intArrayDecl = (IntArrayDecl*)expr;
			return new ArrayVarIntExpr(*intArrayDecl);
		}
		if (expr->type() == Expr::typeDoubleArrayVar)
		{
			DoubleArrayDecl* doubleArrayDecl = (DoubleArrayDecl*)expr;
			return new ArrayVarDoubleExpr(*doubleArrayDecl);
		}
		if (expr->type() == Expr::typeStringArrayVar)
		{
			StringArrayDecl* stringArrayDecl = (StringArrayDecl*)expr;
			return new ArrayVarStringExpr(*stringArrayDecl);
		}
		if (expr->type() == Expr::typeColorArrayVar)
		{
			ColorArrayDecl* colorArrayDecl = (ColorArrayDecl*)expr;
			return new ArrayVarColorExpr(*colorArrayDecl);
		}
		if (expr->type() == Expr::typePointArrayVar)
		{
			PointArrayDecl* pointArrayDecl = (PointArrayDecl*)expr;
			return new ArrayVarPointExpr(*pointArrayDecl);
		}
		error(tree, "Type of var not implemented yet");
	}
	static Ident ident_statements("statements");
	if (tree.isTree(ident_statements))
	{
		VarDecl* statement_context = context;
		AbstractParseTree statements = tree.part(1);
		ASSERT_ELSE(statements.isList()) return new Expr();
		StatementsExpr* stats_expr = new StatementsExpr(statements.nrParts());
		int i = 0;
		for (AbstractParseTree::iterator it(statements); it.more(); it.next())
		{
			Expr* expr = compile(it, statement_context);
			if (expr->type() == Expr::typeNone) 
				return new Expr();
			stats_expr->set(i++, (VoidExpr*)expr);
		}
		return stats_expr;
	}

	static Ident ident_for("for");
	if (tree.isTree(ident_for))
	{
		ASSERT_ELSE(tree.nrParts() == 4 && tree.part(1).isIdent()) return new Expr();
		Ident var = tree.part(1).identName();
		VarIntDecl* varIntDecl = new VarIntDecl(var, context, 0);
		//Context* for_context = new Context(var, varIntDecl, context);
		VarDecl* for_context = varIntDecl;
		Expr* from_expr = compile(tree.part(2), context);
		if (!assertIntType(from_expr, tree.part(2))) return new Expr();
		Expr* to_expr = compile(tree.part(3), context);
		if (!assertIntType(to_expr, tree.part(3))) return new Expr();
		VoidExpr* body = toVoidExpr(compile(tree.part(4), for_context), tree.part(4));
		return new ForIntExpr(varIntDecl, (IntExpr*)from_expr, (IntExpr*)to_expr, body);
	}

	static Ident ident_while("while");
	if (tree.isTree(ident_while))
	{
		ASSERT_ELSE(tree.nrParts() == 2) return new Expr();
		Expr* condition = compile(tree.part(1), context);
		if (!assertBoolType(condition, tree.part(1))) return new Expr();
		VoidExpr* body = toVoidExpr(compile(tree.part(2), context), tree.part(2));
		return new WhileExpr((BoolExpr*)condition, body);
	}

	static Ident ident_if("if");
	if (tree.isTree(ident_if))
	{
		ASSERT_ELSE(tree.nrParts() == 3) return new Expr();
		Expr* condition = compile(tree.part(1), context);
		if (!assertBoolType(condition, tree.part(1))) return new Expr();
		VoidExpr* then_part = toVoidExpr(compile(tree.part(2), context), tree.part(2));
		VoidExpr* else_part = tree.part(3).isEmpty() ? 0 : toVoidExpr(compile(tree.part(3), context), tree.part(3));
		return new IfExpr((BoolExpr*)condition, then_part, else_part);
	}

	static Ident ident_bool("bool");
	static Ident ident_int("int");
	static Ident ident_double("double");
	static Ident ident_string("string");
	static Ident ident_textbuffer("textbuffer");
	static Ident ident_image("image");
	static Ident ident_color("color");
	static Ident ident_point("point");

	static Ident ident_vardecl("vardecl");
	if (tree.isTree(ident_vardecl))
	{
		ASSERT_ELSE(tree.nrParts() == 3 && tree.part(2).isIdent()) return new Expr();
		AbstractParseTree type = tree.part(1);
		Ident ident = tree.part(2).identName();
		Expr* init_expr = 0;
		if (!tree.part(3).isEmpty())
			init_expr = compile(tree.part(3), context);
		VarDecl* decl_expr = 0;
		if (type.isTree(ident_bool))
		{
			if (init_expr != 0 && !assertBoolType(init_expr, tree.part(3))) return new Expr();
			decl_expr = new VarBoolDecl(ident, context, (BoolExpr*)init_expr);
		}
		if (type.isTree(ident_int))
		{
			if (init_expr != 0 && !assertIntType(init_expr, tree.part(3))) return new Expr();
			decl_expr = new VarIntDecl(ident, context, (IntExpr*)init_expr);
		}
		if (type.isTree(ident_double))
		{
			if (init_expr != 0 && !assertDoubleIntType(init_expr, tree.part(3))) return new Expr();
			decl_expr = new VarDoubleDecl(ident, context, init_expr == 0 ? 0 : init_expr->type() == Expr::typeInt ? new IntToDoubleExpr((IntExpr*)init_expr): (DoubleExpr*)init_expr);
		}
		if (type.isTree(ident_string))
		{
			if (init_expr != 0 && !assertStringType(init_expr, tree.part(3))) return new Expr();
			decl_expr = new VarStringDecl(ident, context, init_expr == 0 ? 0 : (StringExpr*)init_expr);
		}
		if (type.isTree(ident_textbuffer))
		{
			if (init_expr != 0) { error(tree.part(3), "Initialization not allowed"); return new Expr(); }
			decl_expr = new VarTextBufferDecl(ident, context);
		}
		if (type.isTree(ident_image))
		{
			if (init_expr != 0) { error(tree.part(3), "Initialization not allowed"); return new Expr(); }
			decl_expr = new VarImageDecl(ident, context);
		}
		if (type.isTree(ident_color))
		{
			if (init_expr != 0 && !assertColorType(init_expr, tree.part(3))) return new Expr();
			decl_expr = new VarColorDecl(ident, context,(ColorExpr*)init_expr);
		}
		if (type.isTree(ident_point))
		{
			if (init_expr != 0 && !assertPointType(init_expr, tree.part(3))) return new Expr();
			decl_expr = new VarPointDecl(ident, context,(PointExpr*)init_expr);
		}

		if (decl_expr == 0)
		{
			error(tree, "Not implemented type");
			return new Expr();
		}
		//context = new Context(ident, decl_expr, context);
		context = decl_expr;

		return decl_expr;
	}

	static Ident ident_arraydecl("arraydecl");
	if (tree.isTree(ident_arraydecl))
	{
		ASSERT_ELSE(tree.nrParts() == 3 && tree.part(2).isIdent()) return new Expr();
		AbstractParseTree type = tree.part(1);
		Ident ident = tree.part(2).identName();
		int nr_dim = 0;
		Expr** dim_exprs = compile_list(tree.part(3), nr_dim, context);
		VarDecl* decl_expr = 0;
		for (int i = 0; i < nr_dim; i++)
			if (!assertIntType(dim_exprs[i], tree.part(3).part(i))) return new Expr();
		if (type.isTree(ident_bool))
			decl_expr = new BoolArrayDecl(ident, context, nr_dim, (IntExpr**)dim_exprs);
		if (type.isTree(ident_int))
			decl_expr = new IntArrayDecl(ident, context, nr_dim, (IntExpr**)dim_exprs);
		if (type.isTree(ident_double))
			decl_expr = new DoubleArrayDecl(ident, context, nr_dim, (IntExpr**)dim_exprs);
		if (type.isTree(ident_string))
			decl_expr = new StringArrayDecl(ident, context, nr_dim, (IntExpr**)dim_exprs);
		if (type.isTree(ident_color))
			decl_expr = new ColorArrayDecl(ident, context, nr_dim, (IntExpr**)dim_exprs);
		if (type.isTree(ident_point))
			decl_expr = new PointArrayDecl(ident, context, nr_dim, (IntExpr**)dim_exprs);

		if (decl_expr == 0)
		{
			error(tree, "Array for this type not implemented yet");
			return new Expr();
		}
		//context = new Context(ident, decl_expr, context);
		context = decl_expr;

		return decl_expr;
	}

	static Ident ident_post_inc("post_inc");
	if (tree.isTree(ident_post_inc))
	{
		ASSERT_ELSE(tree.nrParts() == 1) return new Expr();
		Expr* rhs = compile(tree.part(1), context);
		if (!assertRefIntType(rhs, tree.part(1))) return new Expr();
		return new PostIncExpr((RefIntExpr*)rhs);
	}
	static Ident ident_post_dec("post_dec");
	if (tree.isTree(ident_post_dec))
	{
		ASSERT_ELSE(tree.nrParts() == 1) return new Expr();
		Expr* rhs = compile(tree.part(1), context);
		if (!assertRefIntType(rhs, tree.part(1))) return new Expr();
		return new PostDecExpr((RefIntExpr*)rhs);
	}

	static Ident ident_pre_inc("pre_inc");
	if (tree.isTree(ident_pre_inc))
	{
		ASSERT_ELSE(tree.nrParts() == 1) return new Expr();
		Expr* rhs = compile(tree.part(1), context);
		if (!assertRefIntType(rhs, tree.part(1))) return new Expr();
		return new PreIncExpr((RefIntExpr*)rhs);
	}
	static Ident ident_pre_dec("pre_dec");
	if (tree.isTree(ident_pre_dec))
	{
		ASSERT_ELSE(tree.nrParts() == 1) return new Expr();
		Expr* rhs = compile(tree.part(1), context);
		if (!assertRefIntType(rhs, tree.part(1))) return new Expr();
		return new PreDecExpr((RefIntExpr*)rhs);
	}

	static Ident ident_not("not");
	if (tree.isTree(ident_not))
	{
		ASSERT_ELSE(tree.nrParts() == 1) return new Expr();
		Expr* rhs = compile(tree.part(1), context);
		if (!assertBoolType(rhs, tree.part(1))) return new Expr();
		return new NotBoolExpr((BoolExpr*)rhs);
	}

	static Ident ident_add("add");
	static Ident ident_sub("sub");
	static Ident ident_times("times");
	static Ident ident_div("div");
	static Ident ident_mod("mod");
	static Ident ident_eq("eq");
	static Ident ident_ne("ne");
	static Ident ident_lt("lt");
	static Ident ident_le("le");
	static Ident ident_gt("gt");
	static Ident ident_ge("ge");
	if (   tree.isTree(ident_add)
		|| tree.isTree(ident_sub)
		|| tree.isTree(ident_times)
		|| tree.isTree(ident_div)
		|| tree.isTree(ident_mod)
		|| tree.isTree(ident_eq)
		|| tree.isTree(ident_ne)
		|| tree.isTree(ident_lt)
		|| tree.isTree(ident_le)
		|| tree.isTree(ident_gt)
		|| tree.isTree(ident_ge))
	{
		ASSERT_ELSE(tree.nrParts() == 2) return new Expr();
		Expr* lhs = compile(tree.part(1), context);
		if (!assertDoubleIntColorPointStringType(lhs, tree.part(1))) return new Expr();
		Expr* rhs = compile(tree.part(2), context);
		if (!assertDoubleIntColorPointStringType(rhs, tree.part(2))) return new Expr();
		if (lhs->type() == Expr::typeColor || rhs->type() == Expr::typeColor)
		{
			if (lhs->type() != Expr::typeColor || rhs->type() != Expr::typeColor)
			{
				error(tree, "types do not match");
				return new Expr();
			}
			if (tree.isTree(ident_add))
				return new AddColorExpr((ColorExpr*)lhs, (ColorExpr*)rhs);
			if (tree.isTree(ident_sub))
				return new SubColorExpr((ColorExpr*)lhs, (ColorExpr*)rhs);
			if (tree.isTree(ident_eq))
				return new EqualColorBoolExpr((ColorExpr*)lhs, (ColorExpr*)rhs);
			if (tree.isTree(ident_ne))
				return new UnequalColorBoolExpr((ColorExpr*)lhs, (ColorExpr*)rhs);
		}
		else if (lhs->type() == Expr::typePoint || rhs->type() == Expr::typePoint)
		{
			if (tree.isTree(ident_times))
			{
				if (lhs->type() == Expr::typeInt)
					return new TimesPointExpr(new IntToDoubleExpr((IntExpr*)lhs), (PointExpr*)rhs);
				if (lhs->type() == Expr::typeDouble)
					return new TimesPointExpr((DoubleExpr*)lhs, (PointExpr*)rhs);
				if (rhs->type() == Expr::typeInt)
					return new TimesPointExpr(new IntToDoubleExpr((IntExpr*)rhs), (PointExpr*)lhs);
				if (rhs->type() == Expr::typeDouble)
					return new TimesPointExpr((DoubleExpr*)rhs, (PointExpr*)lhs);
			}
			if (lhs->type() != Expr::typePoint || rhs->type() != Expr::typePoint)
			{
				error(tree, "types do not match");
				return new Expr();
			}
			if (tree.isTree(ident_add))
				return new AddPointExpr((PointExpr*)lhs, (PointExpr*)rhs);
			if (tree.isTree(ident_eq))
				return new EqualPointBoolExpr((PointExpr*)lhs, (PointExpr*)rhs);
			if (tree.isTree(ident_ne))
				return new UnequalPointBoolExpr((PointExpr*)lhs, (PointExpr*)rhs);
		}
		else if (lhs->type() == Expr::typeString || rhs->type() == Expr::typeString)
		{
			if (lhs->type() != Expr::typeString || rhs->type() != Expr::typeString)
			{
				error(tree, "types do not match");
				return new Expr();
			}
			if (tree.isTree(ident_add))
				return new AddStringExpr((StringExpr*)lhs, (StringExpr*)rhs);
			if (tree.isTree(ident_eq))
				return new EqualStringBoolExpr((StringExpr*)lhs, (StringExpr*)rhs);
			if (tree.isTree(ident_ne))
				return new UnequalStringBoolExpr((StringExpr*)lhs, (StringExpr*)rhs);
			if (tree.isTree(ident_lt))
				return new SmallerStringBoolExpr((StringExpr*)lhs, (StringExpr*)rhs);
			if (tree.isTree(ident_le))
				return new SmallerOrEqualStringBoolExpr((StringExpr*)lhs, (StringExpr*)rhs);
			if (tree.isTree(ident_gt))
				return new LargerStringBoolExpr((StringExpr*)lhs, (StringExpr*)rhs);
			if (tree.isTree(ident_ge))
				return new LargerOrEqualStringBoolExpr((StringExpr*)lhs, (StringExpr*)rhs);
		}
		else if (lhs->type() == Expr::typeDouble || rhs->type() == Expr::typeDouble)
		{
			if (lhs->type() == Expr::typeInt)
				lhs = new IntToDoubleExpr((IntExpr*)lhs);
			if (rhs->type() == Expr::typeInt)
				rhs = new IntToDoubleExpr((IntExpr*)rhs);
			if (tree.isTree(ident_add))
				return new AddDoubleExpr((DoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (tree.isTree(ident_sub))
				return new SubDoubleExpr((DoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (tree.isTree(ident_times))
				return new TimesDoubleExpr((DoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (tree.isTree(ident_div))
				return new DivDoubleExpr((DoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (tree.isTree(ident_mod)) { error(tree, "% not allowed for double values"); return new Expr(); }
			if (tree.isTree(ident_eq))
				return new EqualDoubleBoolExpr((DoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (tree.isTree(ident_ne))
				return new UnequalDoubleBoolExpr((DoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (tree.isTree(ident_lt))
				return new SmallerDoubleBoolExpr((DoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (tree.isTree(ident_le))
				return new SmallerOrEqualDoubleBoolExpr((DoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (tree.isTree(ident_gt))
				return new LargerDoubleBoolExpr((DoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (tree.isTree(ident_ge))
				return new LargerOrEqualDoubleBoolExpr((DoubleExpr*)lhs, (DoubleExpr*)rhs);
		}
		else if (lhs->type() == Expr::typeInt && rhs->type() == Expr::typeInt)
		{
			if (tree.isTree(ident_add))
				return new AddIntExpr((IntExpr*)lhs, (IntExpr*)rhs);
			if (tree.isTree(ident_sub))
				return new SubIntExpr((IntExpr*)lhs, (IntExpr*)rhs);
			if (tree.isTree(ident_times))
				return new TimesIntExpr((IntExpr*)lhs, (IntExpr*)rhs);
			if (tree.isTree(ident_div))
				return new DivIntExpr((IntExpr*)lhs, (IntExpr*)rhs);
			if (tree.isTree(ident_mod))
				return new ModIntExpr((IntExpr*)lhs, (IntExpr*)rhs);
			if (tree.isTree(ident_eq))
				return new EqualIntBoolExpr((IntExpr*)lhs, (IntExpr*)rhs);
			if (tree.isTree(ident_ne))
				return new UnequalIntBoolExpr((IntExpr*)lhs, (IntExpr*)rhs);
			if (tree.isTree(ident_lt))
				return new SmallerIntBoolExpr((IntExpr*)lhs, (IntExpr*)rhs);
			if (tree.isTree(ident_le))
				return new SmallerOrEqualIntBoolExpr((IntExpr*)lhs, (IntExpr*)rhs);
			if (tree.isTree(ident_gt))
				return new LargerIntBoolExpr((IntExpr*)lhs, (IntExpr*)rhs);
			if (tree.isTree(ident_ge))
				return new LargerOrEqualIntBoolExpr((IntExpr*)lhs, (IntExpr*)rhs);
		}
	}

	static Ident ident_false("false");
	if (tree.isTree(ident_false))
		return new ConstBoolExpr(false);
	static Ident ident_true("true");
	if (tree.isTree(ident_true))
		return new ConstBoolExpr(true);
	
	static Ident ident_and("and");
	static Ident ident_or("or");
	if (   tree.isTree(ident_and)
		|| tree.isTree(ident_or))
	{
		ASSERT_ELSE(tree.nrParts() == 2) return new Expr();
		Expr* lhs = compile(tree.part(1), context);
		if (!assertBoolType(lhs, tree.part(1))) return new Expr();
		Expr* rhs = compile(tree.part(2), context);
		if (!assertBoolType(rhs, tree.part(2))) return new Expr();
		if (tree.isTree(ident_and))
			return new AndBoolExpr((BoolExpr*)lhs, (BoolExpr*)rhs);
		if (tree.isTree(ident_or))
			return new OrBoolExpr((BoolExpr*)lhs, (BoolExpr*)rhs);
	}

	static Ident ident_if_expr("if_expr");
	if (tree.isTree(ident_if_expr))
	{
		ASSERT_ELSE(tree.nrParts() == 3) return new Expr();
		Expr* condExpr = compile(tree.part(1), context);
		if (!assertBoolType(condExpr, tree.part(1))) return new Expr();
		Expr* thenExpr = compile(tree.part(2), context);
		if (!assertBoolDoubleIntColorPointStringType(thenExpr, tree.part(2))) return new Expr();
		Expr* elseExpr = compile(tree.part(3), context);
		if (!assertBoolDoubleIntColorPointStringType(elseExpr, tree.part(3))) return new Expr();
		if (thenExpr->type() == Expr::typeColor || elseExpr->type() == Expr::typeColor)
		{
			if (thenExpr->type() != Expr::typeColor || elseExpr->type() != Expr::typeColor)
			{
				error(tree, "types of then and else part do not match");
				return new Expr();
			}
			return new IfColorExpr((BoolExpr*)condExpr, (ColorExpr*)thenExpr, (ColorExpr*)elseExpr);
		}
		if (thenExpr->type() == Expr::typePoint || elseExpr->type() == Expr::typePoint)
		{
			if (thenExpr->type() != Expr::typePoint || elseExpr->type() != Expr::typePoint)
			{
				error(tree, "types of then and else part do not match");
				return new Expr();
			}
			return new IfPointExpr((BoolExpr*)condExpr, (PointExpr*)thenExpr, (PointExpr*)elseExpr);
		}
		if (thenExpr->type() == Expr::typeString || elseExpr->type() == Expr::typeString)
		{
			if (thenExpr->type() != Expr::typeString || elseExpr->type() != Expr::typeString)
			{
				error(tree, "types of then and else part do not match");
				return new Expr();
			}
			return new IfStringExpr((BoolExpr*)condExpr, (StringExpr*)thenExpr, (StringExpr*)elseExpr);
		}
		if (thenExpr->type() == Expr::typeBool || elseExpr->type() == Expr::typeBool)
		{
			if (thenExpr->type() != Expr::typeBool || elseExpr->type() != Expr::typeBool)
			{
				error(tree, "types of then and else part do not match");
				return new Expr();
			}
			return new IfBoolExpr((BoolExpr*)condExpr, (BoolExpr*)thenExpr, (BoolExpr*)elseExpr);
		}
		if (thenExpr->type() == Expr::typeDouble || elseExpr->type() == Expr::typeDouble)
		{
			if (thenExpr->type() == Expr::typeInt)
				thenExpr = new IntToDoubleExpr((IntExpr*)thenExpr);
			if (elseExpr->type() == Expr::typeInt)
				elseExpr = new IntToDoubleExpr((IntExpr*)elseExpr);
			return new IfDoubleExpr((BoolExpr*)condExpr, (DoubleExpr*)thenExpr, (DoubleExpr*)elseExpr);
		}
		return new IfIntExpr((BoolExpr*)condExpr, (IntExpr*)thenExpr, (IntExpr*)elseExpr);
	}

	static Ident ident_assignment("assignment");
	if (tree.isTree(ident_assignment))
	{
		ASSERT_ELSE(tree.nrParts() == 3) return new Expr();
		Expr* lhs = compile(tree.part(1), context);
		AbstractParseTree kind = tree.part(2);
		if (!lhs->isRef()) { error(tree.part(1), "Expecting reference"); return new Expr(); }
		Expr* rhs = compile(tree.part(3), context);

		Ident ident_ass("ass");
		Ident ident_times_ass("times_ass");
		Ident ident_div_ass("div_ass");
		Ident ident_mod_ass("mod_ass");
		Ident ident_add_ass("add_ass");
		Ident ident_sub_ass("sub_ass");
		if (lhs->type() == Expr::typeBool)
		{
			if (!assertBoolType(rhs, tree.part(3))) return new Expr();
			if (kind.isTree(ident_ass))
				return new BoolAssignmentExpr((RefBoolExpr*)lhs, (BoolExpr*)rhs);
		}
		if (lhs->type() == Expr::typeInt)
		{
			if (!assertIntType(rhs, tree.part(3))) return new Expr();
			if (kind.isTree(ident_ass))
				return new IntAssignmentExpr((RefIntExpr*)lhs, (IntExpr*)rhs);
			if (kind.isTree(ident_times_ass))
				return new IntTimesAssignmentExpr((RefIntExpr*)lhs, (IntExpr*)rhs);
			if (kind.isTree(ident_div_ass))
				return new IntDivAssignmentExpr((RefIntExpr*)lhs, (IntExpr*)rhs);
			if (kind.isTree(ident_mod_ass))
				return new IntModAssignmentExpr((RefIntExpr*)lhs, (IntExpr*)rhs);
			if (kind.isTree(ident_add_ass))
				return new IntAddAssignmentExpr((RefIntExpr*)lhs, (IntExpr*)rhs);
			if (kind.isTree(ident_sub_ass))
				return new IntSubAssignmentExpr((RefIntExpr*)lhs, (IntExpr*)rhs);
		}
		if (lhs->type() == Expr::typeDouble)
		{
			if (!assertDoubleIntType(rhs, tree.part(3))) return new Expr();
			if (rhs->type() == Expr::typeInt)
				rhs = new IntToDoubleExpr((IntExpr*)rhs);
			if (kind.isTree(ident_ass))
				return new DoubleAssignmentExpr((RefDoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (kind.isTree(ident_times_ass))
				return new DoubleTimesAssignmentExpr((RefDoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (kind.isTree(ident_div_ass))
				return new DoubleDivAssignmentExpr((RefDoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (kind.isTree(ident_mod_ass))
			{	error(kind, "Modulo not defined for double");
				return new Expr();
			}
			if (kind.isTree(ident_add_ass))
				return new DoubleAddAssignmentExpr((RefDoubleExpr*)lhs, (DoubleExpr*)rhs);
			if (kind.isTree(ident_sub_ass))
				return new DoubleSubAssignmentExpr((RefDoubleExpr*)lhs, (DoubleExpr*)rhs);
		}
		if (lhs->type() == Expr::typeString)
		{
			if (kind.isTree(ident_ass))
			{
				if (!assertStringType(rhs, tree.part(3))) return new Expr();
				return new StringAssignmentExpr((RefStringExpr*)lhs, (StringExpr*)rhs);
			}
		}
		if (lhs->type() == Expr::typeColor)
		{
			if (kind.isTree(ident_ass) || kind.isTree(ident_add_ass) || kind.isTree(ident_sub_ass))
			{
				if (!assertColorType(rhs, tree.part(3))) return new Expr();
				if (kind.isTree(ident_ass))
					return new ColorAssignmentExpr((RefColorExpr*)lhs, (ColorExpr*)rhs);
				if (kind.isTree(ident_add_ass))
					return new ColorAddAssignmentExpr((RefColorExpr*)lhs, (ColorExpr*)rhs);
				if (kind.isTree(ident_sub_ass))
					return new ColorSubAssignmentExpr((RefColorExpr*)lhs, (ColorExpr*)rhs);
			}
			if (kind.isTree(ident_times_ass))
			{
				if (!assertDoubleIntType(rhs, tree.part(3))) return new Expr();
				if (rhs->type() == Expr::typeInt)
					rhs = new IntToDoubleExpr((IntExpr*)rhs);
				return new ColorTimesAssignmentExpr((RefColorExpr*)lhs, (DoubleExpr*)rhs);
			}
		}
		if (lhs->type() == Expr::typePoint)
		{
			if (kind.isTree(ident_ass) || kind.isTree(ident_add_ass) || kind.isTree(ident_sub_ass))
			{
				if (!assertPointType(rhs, tree.part(3))) return new Expr();
				if (kind.isTree(ident_ass))
					return new PointAssignmentExpr((RefPointExpr*)lhs, (PointExpr*)rhs);
				if (kind.isTree(ident_add_ass))
					return new PointAddAssignmentExpr((RefPointExpr*)lhs, (PointExpr*)rhs);
			}
			if (kind.isTree(ident_times_ass))
			{
				if (!assertDoubleIntType(rhs, tree.part(3))) return new Expr();
				if (rhs->type() == Expr::typeInt)
					rhs = new IntToDoubleExpr((IntExpr*)rhs);
				return new PointTimesAssignmentExpr((RefPointExpr*)lhs, (DoubleExpr*)rhs);
			}
		}
	}

	static Ident ident_field("field");
	if (tree.isTree(ident_field))
	{
		ASSERT_ELSE(tree.nrParts() == 2 && tree.part(2).isIdent()) return new Expr();
		Ident fieldName = tree.part(2).identName();
		Expr* target = compile(tree.part(1), context);
		if (!target->isRef()) { error(tree.part(1), "Target needs to be reference"); return new Expr(); }
		Expr* result = 0;
		if (target->type() == Expr::typeImage)
		{
			static Ident ident_width("width");
			if (fieldName == ident_width)
				result = new ImageWidthFieldExpr((VarImageExpr*)target);
			static Ident ident_height("height");
			if (fieldName == ident_height)
				result = new ImageHeightFieldExpr((VarImageExpr*)target);
		}
		if (target->type() == Expr::typeColor)
		{
			static Ident ident_r("r");
			if (fieldName == ident_r)
				result = new ColorRedFieldExpr((VarColorExpr*)target);
			static Ident ident_g("g");
			if (fieldName == ident_g)
				result = new ColorGreenFieldExpr((VarColorExpr*)target);
			static Ident ident_b("b");
			if (fieldName == ident_b)
				result = new ColorBlueFieldExpr((VarColorExpr*)target);
		}
		if (target->type() == Expr::typePoint)
		{
			static Ident ident_x("x");
			if (fieldName == ident_x)
				result = new PointXFieldExpr((VarPointExpr*)target);
			static Ident ident_y("y");
			if (fieldName == ident_y)
				result = new PointYFieldExpr((VarPointExpr*)target);
		}
		if (target->type() == Expr::typeTextBuffer)
		{
			static Ident ident_nrLines("nrLines");
			if (fieldName == ident_nrLines)
				result = new TextBufferNrLinesFieldExpr((VarTextBufferExpr*)target);
		}
		if (result == 0)
		{
			error(tree, "unknown field");
			return new Expr();
		}
		return result;
	}

	static Ident ident_function("function");
	if (tree.isTree(ident_function))
	{
		ASSERT_ELSE(tree.nrParts() == 2 && tree.part(1).isIdent()) return new Expr();
		Ident functionName = tree.part(1).identName();
		AbstractParseTree arg_list = tree.part(2);
		int nr_args = 0;
		Expr** arg_exprs = compile_list(arg_list, nr_args, context);
		Expr* result = 0;
		static Ident ident_floor("floor");
		static Ident ident_round("round");
		static Ident ident_sqrt("sqrt");
		static Ident ident_log("log");
		static Ident ident_sin("sin");
		static Ident ident_cos("cos");
		if (   functionName == ident_floor
			|| functionName == ident_round
			|| functionName == ident_sqrt
			|| functionName == ident_log
			|| functionName == ident_sin
			|| functionName == ident_cos)
		{
			if (nr_args != 1) { error(arg_list, "Should have one argument"); return new Expr(); }
			Expr* arg = arg_exprs[0];
			if (!assertDoubleIntType(arg, arg_list)) return new Expr();
			if (arg->type() == Expr::typeInt)
				arg = new IntToDoubleExpr((IntExpr*)arg);
			if (functionName == ident_floor)
				result = new FloorIntExpr((DoubleExpr*)arg);
			else if (functionName == ident_round)
				result = new RoundIntExpr((DoubleExpr*)arg);
			else if (functionName == ident_sqrt)
				result = new SqrtDoubleExpr((DoubleExpr*)arg);
			else if (functionName == ident_log)
				result = new LogDoubleExpr((DoubleExpr*)arg);
			else if (functionName == ident_sin)
				result = new SinDoubleExpr((DoubleExpr*)arg);
			else if (functionName == ident_cos)
				result = new CosDoubleExpr((DoubleExpr*)arg);
			delete arg_exprs;
		}
		static Ident ident_atan2("atan2");
		if (functionName == ident_atan2)
		{
			if (nr_args != 2) { error(arg_list, "Should have two arguments"); return new Expr(); }
			Expr* argx = arg_exprs[0];
			if (!assertDoubleIntType(argx, arg_list)) return new Expr();
			Expr* argy = arg_exprs[1];
			if (!assertDoubleIntType(argy, arg_list)) return new Expr();
			if (argx->type() == Expr::typeInt)
				argx = new IntToDoubleExpr((IntExpr*)argx);
			if (argy->type() == Expr::typeInt)
				argy = new IntToDoubleExpr((IntExpr*)argy);
			result = new Atan2DoubleExpr((DoubleExpr*)argx, (DoubleExpr*)argy);
			delete arg_exprs;
		}
		static Ident ident_abs("abs");
		if (functionName == ident_abs)
		{
			if (nr_args != 1) { error(arg_list, "Should have one argument"); return new Expr(); }
			if (!assertDoubleIntType(arg_exprs[0], arg_list)) return new Expr();
			if (arg_exprs[0]->type() == Expr::typeInt)
				result = new AbsIntExpr((IntExpr*)arg_exprs[0]);
			if (arg_exprs[0]->type() == Expr::typeDouble)
				result = new AbsDoubleExpr((DoubleExpr*)arg_exprs[0]);
			delete arg_exprs;
		}
		static Ident ident_randomVar("randomVar");
		if (functionName == ident_randomVar)
		{
			if (nr_args != 1) { error(arg_list, "Should have one argument"); return new Expr(); }
			if (!assertDoubleType(arg_exprs[0], arg_list)) return new Expr();
			result = new RandomVarDoubleExpr((DoubleExpr*)arg_exprs[0]);
			delete arg_exprs;
		}
		static Ident ident_random("random");
		if (functionName == ident_random)
		{
			if (nr_args != 1) { error(arg_list, "Should have one argument"); return new Expr(); }
			if (!assertIntType(arg_exprs[0], arg_list)) return new Expr();
			result = new RandomIntExpr((IntExpr*)arg_exprs[0]);
			delete arg_exprs;
		}
		static Ident ident_format("format");
		if (functionName == ident_format)
		{
			if (nr_args == 0) { error(arg_list, "Should have at least one argument"); return new Expr(); }
			if (!assertStringType(arg_exprs[0], arg_list)) return new Expr();
			result = new FormatStringExpr(nr_args, arg_exprs);
		}
		static Ident ident_min("min");
		static Ident ident_max("max");
		if (functionName == ident_min || functionName == ident_max)
		{
			if (nr_args != 2) { error(arg_list, "Should have two arguments"); return new Expr(); }
			Expr* lhs = arg_exprs[0];
			if (!assertDoubleIntType(lhs, arg_list)) return new Expr();
			Expr* rhs = arg_exprs[1];
			if (!assertDoubleIntType(rhs, arg_list)) return new Expr();
			if (lhs->type() == Expr::typeDouble || rhs->type() == Expr::typeDouble)
			{
				if (lhs->type() == Expr::typeInt)
					lhs = new IntToDoubleExpr((IntExpr*)lhs);
				if (rhs->type() == Expr::typeInt)
					rhs = new IntToDoubleExpr((IntExpr*)rhs);
				if (functionName == ident_max)
					result = new MaxDoubleExpr((DoubleExpr*)lhs, (DoubleExpr*)rhs);
				if (functionName == ident_min)
					result = new MinDoubleExpr((DoubleExpr*)lhs, (DoubleExpr*)rhs);
			}
			else
			{
				if (functionName == ident_max)
					result = new MaxIntExpr((IntExpr*)lhs, (IntExpr*)rhs);
				if (functionName == ident_min)
					result = new MinIntExpr((IntExpr*)lhs, (IntExpr*)rhs);
			}
			delete arg_exprs;
		}
		static Ident ident_Color("Color");
		if (functionName == ident_Color)
		{
			if (nr_args != 3) { error(arg_list, "Should have three arguments"); return new Expr(); }
			Expr* red_expr = arg_exprs[0];
			if (!assertDoubleIntType(red_expr, arg_list)) return new Expr();
			Expr* green_expr = arg_exprs[1];
			if (!assertDoubleIntType(green_expr, arg_list)) return new Expr();
			Expr* blue_expr = arg_exprs[2];
			if (!assertDoubleIntType(blue_expr, arg_list)) return new Expr();
			if (red_expr->type() == Expr::typeInt)
				red_expr = new IntToDoubleExpr((IntExpr*)red_expr);
			if (green_expr->type() == Expr::typeInt)
				green_expr = new IntToDoubleExpr((IntExpr*)green_expr);
			if (blue_expr->type() == Expr::typeInt)
				blue_expr = new IntToDoubleExpr((IntExpr*)blue_expr);
			result = new ConstructorColorExpr((DoubleExpr*)red_expr, (DoubleExpr*)green_expr, (DoubleExpr*)blue_expr);
			delete arg_exprs;
		}
		static Ident ident_Point("Point");
		if (functionName == ident_Point)
		{
			if (nr_args != 2) { error(arg_list, "Should have two arguments"); return new Expr(); }
			Expr* x_expr = arg_exprs[0];
			if (!assertDoubleIntType(x_expr, arg_list)) return new Expr();
			Expr* y_expr = arg_exprs[1];
			if (!assertDoubleIntType(y_expr, arg_list)) return new Expr();
			if (x_expr->type() == Expr::typeInt)
				x_expr = new IntToDoubleExpr((IntExpr*)x_expr);
			if (y_expr->type() == Expr::typeInt)
				y_expr = new IntToDoubleExpr((IntExpr*)y_expr);
			result = new ConstructorPointExpr((DoubleExpr*)x_expr, (DoubleExpr*)y_expr);
			delete arg_exprs;
		}
		static Ident ident_distance("distance");
		if (functionName == ident_distance)
		{
			if (nr_args != 2) { error(arg_list, "Should have two arguments"); return new Expr(); }
			Expr* p1_expr = arg_exprs[0];
			if (!assertPointType(p1_expr, arg_list)) return new Expr();
			Expr* p2_expr = arg_exprs[1];
			if (!assertPointType(p2_expr, arg_list)) return new Expr();
			result = new PointDistanceDoubleExpr((PointExpr*)p1_expr, (PointExpr*)p2_expr);
			delete arg_exprs;
		}
		if (result == 0)
		{
			error(tree, "Unknown function");
			for (int i = 0; i < nr_args; i++)
				delete arg_exprs[i];
			delete arg_exprs;
			return new Expr();
		}
		return result;
	}

	static Ident ident_method("method");
	if (tree.isTree(ident_method))
	{
		ASSERT_ELSE(tree.nrParts() == 3 && tree.part(2).isIdent()) return new Expr();
		Ident methodName = tree.part(2).identName();
		Expr* target = compile(tree.part(1), context);
		if (!target->isRef()) { error(tree.part(1), "Target needs to be reference"); return new Expr(); }
		AbstractParseTree arg_list = tree.part(3);
		int nr_args = 0;
		Expr** arg_exprs = compile_list(arg_list, nr_args, context);
		Expr* result = 0;
		if (target->type() == Expr::typeTextBuffer)
		{
			static Ident ident_open("open");
			if (methodName == ident_open)
			{
				if (nr_args != 1) { error(arg_list, "Should have one argument"); return new Expr(); }
				if (!assertStringType(arg_exprs[0], arg_list)) return new Expr();
				result = new TextBufferOpenMethodExpr((VarTextBufferExpr*)target, (StringExpr*)arg_exprs[0]);
			}
			static Ident ident_line("line");
			if (methodName == ident_line)
			{
				if (nr_args != 1) { error(arg_list, "Should have one argument"); return new Expr(); }
				if (!assertIntType(arg_exprs[0], arg_list)) return new Expr();
				result = new TextBufferLineMethodExpr((VarTextBufferExpr*)target, (IntExpr*)arg_exprs[0]);
			}
			static Ident ident_clear("clear");
			if (methodName == ident_clear)
			{
				if (nr_args != 0) { error(arg_list, "Should have no arguments"); return new Expr(); }
				result = new TextBufferClearMethodExpr((VarTextBufferExpr*)target);
			}
			static Ident ident_append("append");
			if (methodName == ident_append)
			{
				if (nr_args != 1) { error(arg_list, "Should have one argument"); return new Expr(); }
				if (!assertStringType(arg_exprs[0], arg_list)) return new Expr();
				result = new TextBufferAppendMethodExpr((VarTextBufferExpr*)target, (StringExpr*)arg_exprs[0]);
			}
		}
		if (target->type() == Expr::typeImage)
		{
			static Ident ident_open("open");
			if (methodName == ident_open)
			{
				if (nr_args != 1) { error(arg_list, "Should have one argument"); return new Expr(); }
				if (!assertStringType(arg_exprs[0], arg_list)) return new Expr();
				result = new ImageOpenMethodExpr((VarImageExpr*)target, (StringExpr*)arg_exprs[0]);
			}
			static Ident ident_close("close");
			if (methodName == ident_close)
			{
				if (nr_args != 0) { error(arg_list, "Should have no arguments"); return new Expr(); }
				result = new ImageCloseMethodExpr((VarImageExpr*)target);
			}
			static Ident ident_load("load");
			if (methodName == ident_load)
			{
				if (nr_args != 0) { error(arg_list, "Should have no arguments"); return new Expr(); }
				result = new ImageLoadMethodExpr((VarImageExpr*)target);
			}
			static Ident ident_new("new");
			if (methodName == ident_new)
			{
				if (nr_args != 2) { error(arg_list, "Should have two arguments"); return new Expr(); }
				if (!assertIntType(arg_exprs[0], arg_list.part(1))) return new Expr();
				if (!assertIntType(arg_exprs[1], arg_list.part(2))) return new Expr();
				result = new ImageNewMethodExpr((VarImageExpr*)target, (IntExpr*)arg_exprs[0], (IntExpr*)arg_exprs[1]);
			}
			static Ident ident_getColor("getColor");
			if (methodName == ident_getColor)
			{
				if (nr_args != 2) { error(arg_list, "Should have two argument"); return new Expr(); }
				if (!assertIntType(arg_exprs[0], arg_list.part(1))) return new Expr();
				if (!assertIntType(arg_exprs[1], arg_list.part(2))) return new Expr();
				result = new ImageGetColorMethodExpr((VarImageExpr*)target, (IntExpr*)arg_exprs[0], (IntExpr*)arg_exprs[1]);
			}
			static Ident ident_getInterpolatedColor("getInterpolatedColor");
			if (methodName == ident_getInterpolatedColor)
			{
				if (nr_args == 1)
				{
					Expr* arg1 = arg_exprs[0];
					if (!assertPointType(arg1, arg_list.part(1))) return new Expr();
					result = new ImageGetInterpolatedPointColorMethodExpr((VarImageExpr*)target, (PointExpr*)arg1);
				}
				else if (nr_args == 2)
				{
					Expr* arg1 = arg_exprs[0];
					if (!assertDoubleIntType(arg1, arg_list.part(1))) return new Expr();
					if (arg1->type() == Expr::typeInt)
						arg1 = new IntToDoubleExpr((IntExpr*)arg1);
					Expr* arg2 = arg_exprs[1];
					if (!assertDoubleIntType(arg2, arg_list.part(2))) return new Expr();
					if (arg2->type() == Expr::typeInt)
						arg2 = new IntToDoubleExpr((IntExpr*)arg2);
					result = new ImageGetInterpolatedColorMethodExpr((VarImageExpr*)target, (DoubleExpr*)arg1, (DoubleExpr*)arg2);
				}
				else
				{
					error(arg_list, "Should have one point or two double arguments"); 
					return new Expr();
				}
			}
			static Ident ident_setColor("setColor");
			if (methodName == ident_setColor)
			{
				if (nr_args != 3) { error(arg_list, "Should have three argument"); return new Expr(); }
				if (!assertIntType(arg_exprs[0], arg_list.part(1))) return new Expr();
				if (!assertIntType(arg_exprs[1], arg_list.part(2))) return new Expr();
				if (!assertColorType(arg_exprs[2], arg_list.part(3))) return new Expr();
				result = new ImageSetColorMethodExpr((VarImageExpr*)target, (IntExpr*)arg_exprs[0], (IntExpr*)arg_exprs[1], (ColorExpr*)arg_exprs[2]);
			}
			static Ident ident_getPoint("getPoint");
			if (methodName == ident_getPoint)
			{
				if (nr_args != 1) { error(arg_list, "Should have one argument"); return new Expr(); }
				if (!assertIntType(arg_exprs[0], arg_list.part(1))) return new Expr();
				result = new ImageGetPointMethodExpr((VarImageExpr*)target, (IntExpr*)arg_exprs[0]);
			}
		}
		if (result == 0)
		{
			error(tree, "Unknown method");
			for (int i = 0; i < nr_args; i++)
				delete arg_exprs[i];
			delete arg_exprs;
			return new Expr();
		}
		delete arg_exprs;
		return result;
	}

	static Ident ident_array("array");
	if (tree.isTree(ident_array))
	{
		ASSERT_ELSE(tree.nrParts() == 2) return new Expr();
		Expr* target = compile(tree.part(1), context);
		if (target->nrDimensions() == 0) { error(tree.part(1), "Target needs to be array variable"); return new Expr(); }
		AbstractParseTree arg_list = tree.part(2);
		if (target->nrDimensions() != arg_list.nrParts()) { error(tree.part(1), "Nr of dimensions does not match"); return new Expr(); }
		int nr_args = 0;
		Expr** arg_exprs = compile_list(arg_list, nr_args, context);
		if (nr_args != target->nrDimensions()) { error(tree.part(1), "Number dimensions does not match"); return new Expr(); }
		for (int i = 0; i < nr_args; i++)
		{
			if (!assertDoubleIntType(arg_exprs[i], arg_list.part(i))) return new Expr();
			if (arg_exprs[i]->type() == Expr::typeDouble)
				arg_exprs[i] = new RoundIntExpr((DoubleExpr*)arg_exprs[i]);
		}
		Expr* result = 0;
		if (target->type() == Expr::typeBool)
			result = new ArrayDerefBoolExpr((ArrayVarBoolExpr*)target, nr_args, (IntExpr**)arg_exprs);
		if (target->type() == Expr::typeInt)
			result = new ArrayDerefIntExpr((ArrayVarIntExpr*)target, nr_args, (IntExpr**)arg_exprs);
		if (target->type() == Expr::typeDouble)
			result = new ArrayDerefDoubleExpr((ArrayVarDoubleExpr*)target, nr_args, (IntExpr**)arg_exprs);
		if (target->type() == Expr::typeColor)
			result = new ArrayDerefColorExpr((ArrayVarColorExpr*)target, nr_args, (IntExpr**)arg_exprs);
		if (target->type() == Expr::typePoint)
			result = new ArrayDerefPointExpr((ArrayVarPointExpr*)target, nr_args, (IntExpr**)arg_exprs);

		if (result == 0)
		{
			error(tree, "Array for given type not supported");
			for (int i = 0; i < nr_args; i++)
				delete arg_exprs[i];
			delete arg_exprs;
			return new Expr();
		}
		return result;
	}

	error(tree, "Not parsed yet");

	return new Expr();
}

Expr** compile_list(AbstractParseTree arg_list, int &nr_args, VarDecl* &context)
{
	nr_args = arg_list.nrParts();
	Expr** arg_exprs = 0;
	if (nr_args > 0)
	{
		arg_exprs = new Expr*[nr_args];
		for (int i = 0; i < nr_args; i++)
			arg_exprs[i] = compile(arg_list.part(i+1), context);
	}
	return arg_exprs;
}

/*
	Initialization of the IParse grammar
	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	Below the procedure that initializes the Abstract Program Tree
	for the input grammar of IParse itself is given. This was
	generated by calling the procedure print_tree_to_c above.
*/

#include "ScriptGrammar.h"

void executeScript(const char* script, CCrystalTextView* view)
{
	CWaitCursor waitCursor;
	currentView = view;

	IParser* parser = new IParser();
	BasicScanner* scanner = new BasicScanner();
	parser->setScanner(scanner);
	parser->setDebugLevel(false, false, false);
	AbstractParseTree script_grammar_tree;
	init_ScriptGrammar(script_grammar_tree);
	parser->loadGrammer(script_grammar_tree);

	AbstractParseTree parse_tree;
	if (!parser->parse(script, "root", parse_tree))
	{
		TRACE("Failed");
		parser->traceExpected();
		char buffer[1000];
		buffer[0] = '\0';
		int len = 0;
		for (AbstractParser::ExpectAccessor expectAccessor(parser); expectAccessor.more(); expectAccessor.next())
		{
			if (len < 900)
			{
				sprintf(buffer+len, " %s", expectAccessor.symbol());
				len += strlen(buffer+len);
			}
		}
		long line;
		long column;
		parser->getExpectedPos(line, column);
		CString msg;
		msg.Format("Expected%s", buffer);
		error(line, column, msg);
		/*setCursorAt(line-1, column-1);
		CString prompt;
		prompt.Format("%d,%d: Expected%s", line, column, buffer);
		AfxMessageBox(prompt);
		CMainFrame* pMainFrame = (CMainFrame*) AfxGetApp()->m_pMainWnd;
		if (pMainFrame != 0)
			pMainFrame->UpdateStatus(msg);*/
		currentView = 0;
		return;
	}

	VarDecl* context = 0;
	reported_error = false;
	Ptr<Expr> statements = compile(parse_tree, context);
	if (reported_error)
		return;
	//TRACE1("total expr = %d\n", total_nr_expr);
	if (statements != 0 && statements->type() == Expr::typeVoid)
		((VoidExpr*)(Expr*)statements)->eval();
	currentView = 0;
}